/* ---------------------------------------------------------------- */
/* Copyright (c) Atelier de BioInformatique			    */
/* @file: angle_io.c						    */
/* @desc: angle sequence IO in PHIPSI format			    */
/*								    */
/* @history:							    */
/* @+	    <jompo> : Jul 02 : first version			    */
/* ---------------------------------------------------------------- */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mtd.h"	/* for INDX	*/
#include "libmtd_angle_io.h"

#define SECURE   1	/* 1 : secure copy ; 0 : quickest copy 	*/

#define READ_NEXT FALSE
#define PUSH_BACK TRUE

#define SERIAL    TRUE
#define INDEXED   FALSE

#ifdef MAC_OS_C
#define LINE_FEED '\r'
#else
#define LINE_FEED '\n'
#endif

static SBool sRetained = FALSE;

/* -------------------------------------------- */
/* @static: lecture bufferisee			*/
/* -------------------------------------------- */
static char * sA_sNextIOBuffer(FILE *streamin, SBool retain, SBool serial)
{
	char   *buf, *end;
	SBool   reused;
	
	static char sBuffer[BUFSIZ];	/* BUFSIZ in <stdio.h>	*/
	
	reused = (retain || sRetained) && serial;

	sRetained = retain;

	buf = (   reused 
		? sBuffer 
		: fgets(sBuffer, sizeof(sBuffer), streamin));	 	

	if (! buf)		  /* end of file	    */
	    return NULL;

	end = buf + strlen(buf) - 1;

	if (*end == LINE_FEED)	  /* remove trailing \n	    */
	   *end = '\000';

	return (  *buf || reused
	        ? buf
		: sA_sNextIOBuffer(streamin, retain, serial));
}



/* -------------------------------------------- */
/* skip to next space in buffer			*/
/* -------------------------------------------- */
static char *sA_NextSpace(char *buffer)
{
	for (; *buffer ; buffer++)
	   if (isspace(*buffer))
	   	return buffer;
	
	return NULL;
}

/* -------------------------------------------- */
/* returns sequence name (FASTA)		*/
/* -------------------------------------------- */
char *GetAngleName(char *buffer)
{
	char c[2];
	static char name[ANGLE_NAMLEN];

	*c = buffer[ANGLE_NAMLEN];
	 buffer[ANGLE_NAMLEN] = '\000';
	
	if (sscanf(buffer + 1, "%s", name) != 1)
	    (void) strcpy(name, "<no Name>");

	buffer[ANGLE_NAMLEN] = *c;
	 
	return name;
}

/* -------------------------------------------- */
/* returns sequence comment (angle FASTA format)*/
/* -------------------------------------------- */
char *GetAngleComment(char *buffer)
{
	char   c[2], *space;
	static char comment[ANGLE_COMLEN];

	*c = buffer[ANGLE_COMLEN];
	buffer[ANGLE_COMLEN] = '\000';
	
	space = sA_NextSpace(buffer);
	
	(void) strcpy(comment, (space ? space + 1 : "<no comment>"));

	buffer[ANGLE_COMLEN] = *c;

	return comment;
}

/* -------------------------------------------- */
/* liberation d'une sequence			*/
/* -------------------------------------------- */
AngleSequencePtr FreeAngleSequence(AngleSequencePtr seq)
{
	if (seq) {
	    if (seq->seq)  FREE(seq->seq);
	    FREE(seq);
	}

	return NULL;
}
	
/* -------------------------------------------- */
/* allocation d'une sequence			*/
/* -------------------------------------------- */
AngleSequencePtr NewAngleSequence(void)
{
	AngleSequencePtr seq;
	
	if (! (seq = NEW(AngleSequence)))
	    return NULL;
	   
	seq->length   = 0;
	seq->offset   = -1;

	if (! (seq->seq = NEWN(INDX,  BUFSIZ)))
	    return FreeAngleSequence(seq);

	seq->bufsize = BUFSIZ;

	*(seq->name)    = '\000';
	*(seq->comment) = '\000';

	seq->ok = TRUE;
	
	return seq;
}

/* -------------------------------------------- */
/* Nettoyage d'une sequence			*/
/* pour enlever les angles 500.0 (BAD_ANGLE)	*/
/* Note: ils sont superieur a angmax		*/
/* -------------------------------------------- */
void CleanAngleSequence(AngleSequencePtr seq, INDX angmax)
{
   int		i, clean;
   INDX		*s, *ss;

   s = ss = seq->seq;
   clean = 0;
   i = 0;

   while (i < seq->length) {
      if (*s < angmax) {
	*ss = *s;
	ss++;
	s++;
      }
      else {
	/* DEBUG */
	/* fprintf(stderr,"angle %4d cleaned->%3d\n",i+1, (int)*s);  */
	s++;
	clean++;
      }
      i++;
   }

   seq->length -= clean;

   return;
}

/* -------------------------------------------- */
/* lecture/redimensionnement d'une sequence au	*/
/* format Angle					*/
/* <Lecture Serie>				*/
/* returns : FALSE -> last sequence		*/
/*	     TRUE -> more to read		*/
/*           <but> you must check seq->ok !	*/
/*	     that may be 'FALSE' (memory error)	*/
/* -------------------------------------------- */
SBool ReadAngleSequence(FILE *streamin, AngleSequencePtr seq, INDX angmax)
{
				/* angmax is the max in range we	*/
				/* want, i.e. angle are in the 		*/
				/* [0,angmax[ range			*/
	int 	buflen;
	char 	*buffer, namres[256],numcar;
	INDX	*tbuf;
	int	numres;
	float	alpha,tau,distca;

	seq->ok = FALSE;			/* assume error		*/

	buflen = seq->length = 0; 
	
	seq->offset = ftell(streamin);

	buffer = sA_sNextIOBuffer(streamin, READ_NEXT, SERIAL);

	/* OLD if (! (buffer && (*buffer == '>'))) 	 */
						/* sync error		*/
	if (! (buffer && (! strncmp(buffer,"HEADER",6) || ! strncmp(buffer,"PATH", 4) || ! strncmp(buffer,"NBRES",5))))
	    return FALSE;			/* last sequence	*/
	
	if (seq->offset)
	    seq->offset -= (strlen(buffer) + 1);

	(void) strcpy(seq->name,    GetAngleName(buffer));
	
	(void) strcpy(seq->comment, GetAngleComment(buffer));
	
	while ((buffer = sA_sNextIOBuffer(streamin, READ_NEXT, SERIAL))) {

	    /* if (*buffer == '>') { OLD */
	    if (! strncmp(buffer,"HEADER",6) || ! strncmp(buffer,"PATH", 4) || ! strncmp(buffer,"NBRES",5) ) {
	    	(void) sA_sNextIOBuffer(streamin, PUSH_BACK, SERIAL); /* push it back */
		break;
	    }

	    buflen++;
   
	    if (buflen >= seq->bufsize) {
	    
	    	if (! (tbuf = REALLOC(INDX, seq->seq, 2 * buflen + 1)))
	    	   return TRUE;		/* but seq->ok is FALSE	*/

		seq->seq = tbuf;
		
		seq->bufsize = 2 * buflen + 1;
		
	    }		
	    
	    if (sscanf(buffer,"%s %d%c %f %f %f ",
		namres,&numres,&numcar, &alpha,&tau,&distca) != 6) {
		   seq->seq[seq->length] = EOINDX;
	           return TRUE;		/* but seq->ok is FALSE */
	    }
				/* alpha is in range [-180.0,180.0[	*/
				/* except for BAD_ANGLE which is 500.0	*/
				/* we let it as 500.0*angmax/360.0	*/
	    *(seq->seq + seq->length) =  (INDX)(((alpha + 180.0)*(float)angmax)/360.0);
	
	    seq->length++;
	
	}

	seq->seq[seq->length] = EOINDX;

	return (seq->ok = TRUE);
}

/* -------------------------------------------- */
/* lecture/redimensionnement d'une sequence au	*/
/* format Angle					*/
/* <Lecture Indexee>				*/
/* returns : FALSE -> memory error		*/
/*	     TRUE -> read ok			*/
/* -------------------------------------------- */
SBool GetAngleSequence(FILE *streamin, AngleSequencePtr seq, INDX angmax)
{
	int 	buflen;
	char 	*buffer, namres[256],numcar;
	INDX	*tbuf;
	int	numres;
	float	alpha,tau,distca;

	seq->ok = FALSE;			/* assume error		*/

	buflen = seq->length = 0; 
	
	(void) fseek(streamin, seq->offset, SEEK_SET);

	buffer = sA_sNextIOBuffer(streamin, READ_NEXT, INDEXED);

	/* OLD if (! (buffer && (*buffer == '>'))) */ 	/* sync error		*/
	if (! strncmp(buffer,"HEADER",6) || ! strncmp(buffer,"PATH", 4) || ! strncmp(buffer,"NBRES",5) )
	    return FALSE;			/* sync error, last sequence	*/
	
	if (seq->offset)
	    seq->offset -= (strlen(buffer) + 1);

	(void) strcpy(seq->name,    GetAngleName(buffer));
	
	(void) strcpy(seq->comment, GetAngleComment(buffer));
	
	while ((buffer = sA_sNextIOBuffer(streamin, READ_NEXT, INDEXED))) {

	    /* OLD if (*buffer == '>') */
	    if (! strncmp(buffer,"HEADER",6) || ! strncmp(buffer,"PATH", 4) || ! strncmp(buffer,"NBRES",5) )
		break;

	    buflen++;
   
	    if (buflen >= seq->bufsize) {
	    
	    	if (! (tbuf = REALLOC(INDX, seq->seq, 2 * buflen + 1)))
	    	   return FALSE;			

		seq->seq = tbuf;
		
		seq->bufsize = 2 * buflen + 1;
		
	    }		
	
	    seq->length = buflen;
	
	}

	if (sscanf(buffer,"%s %d%c %f %f %f ",
		namres,&numres,&numcar, &alpha,&tau,&distca) != 5) {
		   seq->seq[seq->length] = EOINDX;
	           return TRUE;		/* but seq->ok is FALSE */
	}
				/* alpha is in range [-180.0,180.0[	*/
				/* set it in [0,angmax[ range		*/
	*(seq->seq + seq->length) =  (INDX)(((alpha + 180.0)*(float)angmax)/360.0);
	seq->seq[seq->length] = EOINDX;

	return (seq->ok = TRUE);
}

/* -------------------------------------------- */
/* ecriture d'une sequence au format Angle	*/
/* -------------------------------------------- */
void WriteAngleSequence(FILE *streamou, AngleSequencePtr seq, INDX angmax)
{
	int  i;

	(void) fputs("HEADER ", streamou);
	(void) fputs((*(seq->name)    ? seq->name    : "<no name>")   , streamou);
	(void) fputc(' ', streamou);
	(void) fputs((*(seq->comment) ? seq->comment : "<no comment>"), streamou);
	(void) fputc(LINE_FEED, streamou);


	for (i = 0; i < seq->length; i++) {
		fprintf(streamou, "%3s %5d %7.1f %7.1f %7.1f\n",
				"DUM",i+1,(float)(((seq->seq[i]) * 360.0)/(float)angmax)-180.0,500.0,3.8);
	}

}

/* -------------------------------------------- */
/* rewind db file				*/
/* -------------------------------------------- */
void RewindAngleDB(FILE *streamin)
{
    sRetained = FALSE;	/* forget previous buffer  */
    
    if (streamin)
	rewind(streamin);
}
