/* -------------------------------------------- */
/* @file: mtd_modelio.c	 			*/
/* $Id: libmtd_modelio.c,v 1.2 2005/03/15 11:47:04 annela Exp $ 					*/
/* -------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mtd.h"


/* -------------------------------------------- */
/* copy model src in dest			*/
/* -------------------------------------------- */
void ModelCopy(MTDModel *dest, MTDModel *src) {

   int i;

   if (dest == NULL) {
      fprintf(stderr, "ModelCopy: destination model is NULL\n");
      return;
   }

   memcpy(dest->states, src->states, sizeof(INDX)*(size_t)(src->nstates+1));

   memcpy(dest->phi, src->phi, sizeof(REAL)*(size_t)(src->order));

   for (i = 0; i < src->nstates; i++)
      memcpy(dest->q[i], src->q[i],sizeof(REAL)*(size_t)(src->nstates));

   return;

}

/* -------------------------------------------- */
/* new model					*/
/* -------------------------------------------- */
MTDModel *NewModel(int order, int nstates) {

  MTDModel	*m;
  int		i,j;

  if (! (m = (MTDModel *)malloc(sizeof(MTDModel)))) {
    fprintf(stderr,"NewModel: not enough memory for MTDModel\n");
    return NULL;
  }

  m->order = order;
  m->nstates = nstates;

  if (! (m->phi = (REAL *)malloc((size_t)order*sizeof(REAL)))) {
     free(m);
     return NULL;
  }

  if (! (m->states = (INDX *)malloc((size_t)(nstates+1)*sizeof(INDX)))) {
     free(m->phi);
     free(m);
     return NULL;
  }

  if (! (m->q = (REAL **)malloc((size_t)nstates*sizeof(REAL *)))) {
     free(m->states);
     free(m->phi);
     free(m);
     return NULL;
  }

  for (i = 0; i < nstates; i++) {
     if (! (m->q[i] = (REAL *)malloc((size_t)nstates*sizeof(REAL)))) {
	for (j = 0; j < i; j++)
	   free(m->q[j]);
        free(m->states);
	free(m->phi);
	free(m);
	return NULL;
     }
  }

  return m;

}


/* -------------------------------------------- */
/* free model					*/
/* -------------------------------------------- */
void FreeModel(MTDModel *m) {

  int		i;

  if (m) {
     if (m->phi)
        free(m->phi);
     if (m->q) {
        for (i = 0; i < m->nstates; i++)
	   if (m->q[i])
	      free(m->q[i]);
	free(m->q);
     }
     free(m);
  }
  return;

}


/* -------------------------------------------- */
/* write model					*/
/* -------------------------------------------- */

void writeModel(FILE *stream, MTDModel *model) {

  int i, j;

  (void) fprintf(stream, "%d %d\n", model->order, model->nstates);

#ifndef ANGLES
  (void) fprintf(stream, "%s\n", model->states);
#else
  for (i = 0 ; i < model->nstates ; i++) {
     (void) fprintf(stream, "%4d\n", (int)(model->states[i]));
  }
#endif

  for (i = 0 ; i < model->order ; i++)
     (void) fprintf(stream, "%6.4f ", (float) model->phi[i]);
  (void) fprintf(stream, "\n");

  for (i = 0 ; i < model->nstates ; i++) {
    for (j = 0 ; j < model->nstates ; j++)
       (void) fprintf(stream, "%6.4f ", (float) model->q[i][j]);
    (void) fprintf(stream, "\n");
  }
}

/* -------------------------------------------- */
/* read model					*/
/* -------------------------------------------- */

MTDModel *readModel(FILE *stream) {

  int i, j, n, n0, order,nstates;
  float x;
  char	line[256];
  MTDModel *pmodel;

#ifdef ANGLES
  int state;
#endif

  if (fgets(line, 255, stream)== NULL) {
     fprintf(stderr, "Cannot read first line in input\n");
     return NULL;
  }
  else {
    if ((n = sscanf(line, "%d%d", &order, &nstates)) != 2) {
       fprintf(stderr, "Cannot read order and nstates\n  in line: <%s>\n",line);
       return NULL;
    }
  }

  if (!(pmodel=NewModel(order, nstates))) {
     fprintf(stderr, "readModel::Cannot allocate for model\n");
     return NULL;
  }

#ifndef ANGLES
  if (fgets(line, 255, stream)== NULL) {
     fprintf(stderr, "Cannot read second line in input\n");
     FreeModel(pm);
     return NULL;
  }
  else {
     if ((m = sscanf(line, "%s", pmodel->states)) != 1) {
        fprintf(stderr, "Cannot read states\n");
        FreeModel(pm);
        return NULL;
     }
     n += m;
  }
#else
  for (i = 0 ; i < pmodel->nstates ; i++) {
     if (fgets(line, 255, stream)== NULL) {
       fprintf(stderr, "Cannot read line %d in input\n", i+1);
     }
     if (sscanf(line, "%d", &state) != 1)
       fprintf(stderr, "Cannot read state %d in input line <%s>\n", i+1, line);
     else
	pmodel->states[i] = (INDX)state;
  }
  n ++;
#endif

  for (i = 0 ; i < pmodel->order ; i++) {
    n += fscanf(stream, "%g", &x);
    pmodel->phi[i] = x;
  }

  for (i = 0 ; i < pmodel->nstates ; i++) {
    for (j = 0 ; j < pmodel->nstates ; j++) {
      n += fscanf(stream, "%g", &x);
      pmodel->q[i][j] = x;
    }
  }

  n0 = (pmodel->nstates * pmodel->nstates) + pmodel->order + 3;

  if (n != n0) {
     FreeModel(pmodel);
     pmodel = NULL;
  }

  return pmodel;

}


/* -------------------------------------------- */
/* New set of models				*/
/* -------------------------------------------- */
static ModelSet *NewModelSet(int nmodel) {

  ModelSet *set;

  if (!(set = (ModelSet *)malloc(sizeof(ModelSet)))) {
     fprintf(stderr, "NewModelSet:: Cannot allocate for set\n");
     return NULL;
  }

  set->nmodel = nmodel;

  if (!(set->lprior = (REAL *)malloc(set->nmodel * sizeof(REAL)))) {
     fprintf(stderr, "NewModelSet:: Cannot allocate for set\n");
     free(set);
     return NULL;
  }

  if (!(set->model = (MTDModel **)malloc(set->nmodel * sizeof(MTDModel *)))) {
     fprintf(stderr, "NewModelSet:: Cannot allocate for set\n");
     free(set->lprior);
     free(set);
     return NULL;
  }

  memset(set->model, 0, (size_t)(set->nmodel)*sizeof(MTDModel *));
  memset(set->lprior, 0, (size_t)(set->nmodel)*sizeof(REAL));

  return set;
}

/* -------------------------------------------- */
/* New set of models				*/
/* -------------------------------------------- */
void FreeModelSet(ModelSet *set) {
   int i;

   if (set) {
      for (i = 0; i < set->nmodel; i++) 
         if (set->model[i])
	    FreeModel(set->model[i]);
      if (set->lprior)
	 free(set->lprior);
      free(set);
   }
   return;
}

/* -------------------------------------------- */
/* read set of models				*/
/* -------------------------------------------- */

ModelSet *readModelSet(FILE *stream) {

  int		i, n, nmodel;
  float		x;
  char		line[256];
  ModelSet	*set;


  if (fscanf(stream, "%d", &nmodel) != 1)
    return NULL;

  set = NewModelSet(nmodel);

  for (i = n = 0; i < set->nmodel ; i++) {
    if (fscanf(stream, "%f", &x) != 1) {
       fprintf(stderr, "readModelSet::Cannot read %d-th value\n",i+1);
    }
    n++;
    set->lprior[i] = LOG(x);
#if 0
    fprintf(stderr, "%g %g\n", x, (float) set->lprior[i]);
#endif
  }

			/* just to finish line	*/
  if (fgets(line, 255, stream)== NULL) {
     fprintf(stderr, "readModelSet::Cannot read line in input\n");
     FreeModelSet(set);
     return NULL;
  }

  if (n != set->nmodel) {
    FreeModelSet(set);
    return NULL;
  }

  for (i = 0; i < set->nmodel ; i++) {
    if (!(set->model[i] = readModel(stream))) {
      FreeModelSet(set);
      return NULL;
    }
  }

  return set;
}
