/*
 * This program is for 3D superposition of sets of coordinates.
 * It is based on the quaternions, as proposed by:
 * Zuker & Somorjai, Bulletin of Mathematical Biology,
 * vol. 51, No 1, p 55-78, 1989.
 * 
 * The search for the largest eigen value is based on a QR, and not
 * on the Jacobian based solvation approach used by most programmers.
 * 
 * In our experience, the routine is very robust.
 *
 * 
 * If you want it directly, the main function is:
 * 
 * zuker_superpose(COORDS *c1, COORDS *c2, int len, DtMatrix4x4 M)
 *
 * This program can be freely used, modified and distributed, given that this 
 * header is maintained.
 *
 * If using this routine for publication, please cite F. Guyon and P. Tuffery.
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>             /* pour memset */

#include "libqbestfit.h"

/**************************************************/
/*           Define and structures from QBestFit  */
/**************************************************/

#define DtFloat		double	        /* Le type flottant.	        */

/*typedef DtFloat         DtPoint3[3]; remplac par COORDS*/
typedef DtFloat         DtPoint4[4];
typedef DtFloat         DtMatrix3x3[3][3];


#define boucle(i,min,max) for((i)=(min);(i)<(max);(i)++)
#define max(a,b) (a<b?b:a)
#define min(a,b) (a<b?a:b)
#define sign(a,b) ((b)>=0.0 ? fabs(a) : -fabs(a))
#define EPS 1.e-10

#ifndef MYFREE
#define MYFREE(x)    if ((x)) free(x)
#endif

#define CALLOCY(var, typ, dim)  {var=(typ*)malloc((size_t)(dim)*sizeof(typ));\
                              if((var)==NULL){\
                                  fprintf(stderr,"pb malloc avec %s (%d %s)\n",\
                                  #var,__LINE__,__FILE__);exit(99);}}

#define NEWY(var, typ, dim)  {var=(typ*)malloc((size_t)(dim)*sizeof(typ));\
                              if((var)==NULL){\
                                  fprintf(stderr,"pb malloc avec %s (%d %s)\n",\
                                  #var,__LINE__,__FILE__);exit(99);}}

/**************************************************/
/*            INTERNAL FUNCTION PROTOTYPES        */
/**************************************************/
static double *random_vect(int n);
static double inner(double *x, double *y, int n);
static double *product(double **A, double *x, int n);
static int lu_c (double a[4][4],  int n);
static void resol_lu(double a[4][4], double *b, int n);
static int power(double *a[], int n, int maxiter, double eps, double *v, double *w);
static double best_shift(double *a[], int n);
static double lmax_estim (double a[4][4], int n);
static int shift_power(double *a, int n, int maxiter, double eps, double *v, double *w);
static int inverse_power(double a[4][4], int n, int maxiter, double eps, double *v, double *w);
static int largestEV4(double R[4][4], double v[4], double *vp);
static DtMatrix3x3 *XYCov(DtMatrix3x3 *pM, 
			  COORDS *X, 
			  COORDS *Y, 
			  COORDS *Xmean, 
			  COORDS *Ymean, 
			  int aSze);
static void MkTrnsIIMat4x4(DtMatrix4x4 m, COORDS tr);
static void mulMat4x4(DtMatrix4x4 a,DtMatrix4x4 b,DtMatrix4x4 c);
static double **alloc_mat(int n, int m);
static void free_mat(double **mat, int n);


/**************************************************/
/*                   FUNCTIONS                    */


/**************************************************/
/* * ---------- Transformation d'une coordonnee par rmat. ----------
 * =============================================================== */

COORDS crdRotate(COORDS p,DtMatrix4x4 rmat)
{
  COORDS p2;

  p2.x = p.x * rmat[0][0] + p.y * rmat[1][0] + p.z * rmat[2][0] + rmat[3][0]; 
  p2.y = p.x * rmat[0][1] + p.y * rmat[1][1] + p.z * rmat[2][1] + rmat[3][1]; 
  p2.z = p.x * rmat[0][2] + p.y * rmat[1][2] + p.z * rmat[2][2] + rmat[3][2]; 

  return(p2);

}

#define SQUARED_DISTANCE(K,R) ((K).x - (R).x) * ((K).x - (R).x) + ((K).y - (R).y) * ((K).y - (R).y) + ((K).z - (R).z) * ((K).z - (R).z)

/* double squared_distance(COORDS R,COORDS K)
{
  return (double) ((K.x - R.x) * (K.x - R.x) +
		    (K.y - R.y) * (K.y - R.y) +
		    (K.z - R.z) * (K.z - R.z));
}
*/

/**************************************************/
void print_matrice4x4(DtMatrix4x4 a, char *name, FILE *pOut){
  int i,j;

  for (i=0; i<4; i++) {
    for (j=0; j<4; j++) 
      fprintf (pOut, "%s[%d,%d]=%f ", name, i, j, a[i][j]);
    fprintf (pOut, "\n");
  }
}

/**************************************************/
void print_vector(double *a, int n, char *name, FILE *pOut){
  int i;

  for (i=0; i<n; i++) 
    fprintf (pOut, "%s[%d]=%lf\n ", name, i, a[i]);
}


/**************************************************/
/* 
 * Best fit matrix as proposed by:
 * Zuker & Somorjai, Bulletin of Mathematical Biology,
 * vol. 51, No 1, p 55-78, 1989.
 * 
 * c1, c2 two sets of coordinates to superpose.
 * len : the number of coordiantes of c1, c2.
 *
 * M, the 4x4 transforamtion matrix to pass from c2 to c2 superposed onto c1.
 *
 * Input: 
 *   c1, c2, and len must be given. 
 *   M should be passed, but is meaningless.
 * 
 * Output:
 *   c2 is superposed on c1.
 *   M, a transformation matrix ready for use.
 *
 */
/**************************************************/
double zuker_superpose(COORDS *c1, COORDS *c2, int len, DtMatrix4x4 M)
{
  DtMatrix3x3  C;
  DtMatrix3x3 *pC; 
  DtMatrix4x4  RM;
/*  DtMatrix4x4  TM;*/
  DtMatrix4x4  TMP;
  DtMatrix4x4  TX;
  DtMatrix4x4  TY;
  DtMatrix4x4  P;
/*  DtMatrix4x4  Pd;*/
/*  DtPoint4     lambdas;*/
  DtPoint4     V;
  COORDS     bc1, bc2;
  COORDS     try, p2;
  
  double eval;
  double squared_rms = 0.0L;

  int nCycles;
  int aDot;
/*  fprintf(stderr, "*************************DANS zuker_superpose\n");*/

  /* Compute transformation matrix as proposed by zuker */
  pC = &C;
  pC = XYCov(pC, (COORDS *) c1, (COORDS *) c2, &bc1, &bc2, len);

  P[0][0] = -C[0][0]+C[1][1]-C[2][2];
  P[0][1] = P[1][0] = -C[0][1]-C[1][0];
  P[0][2] = P[2][0] = -C[1][2]-C[2][1];
  P[0][3] = P[3][0] =  C[0][2]-C[2][0];

  P[1][1] = C[0][0]-C[1][1]-C[2][2];
  P[1][2] = P[2][1] = C[0][2]+C[2][0];
  P[1][3] = P[3][1] = C[1][2]-C[2][1];

  P[2][2] = -C[0][0]-C[1][1]+C[2][2];
  P[2][3] = P[3][2] = C[0][1]-C[1][0];

  P[3][3] = C[0][0]+C[1][1]+C[2][2];


/*  print_matrice4x4(P, "zuker P", stderr); */


  nCycles = largestEV4(P, V, &eval);

  RM[0][0] = -V[0]*V[0]+V[1]*V[1]-V[2]*V[2]+V[3]*V[3];
  RM[1][0] =  2*(V[2]*V[3]-V[0]*V[1]);
  RM[2][0] =  2*(V[1]*V[2]+V[0]*V[3]);
  RM[3][0] =  0.;

  RM[0][1] = -2*(V[0]*V[1]+V[2]*V[3]);
  RM[1][1] = V[0]*V[0]-V[1]*V[1]-V[2]*V[2]+V[3]*V[3];
  RM[2][1] =  2*(V[1]*V[3]-V[0]*V[2]);
  RM[3][1] =  0.;

  RM[0][2] =  2*(V[1]*V[2]-V[0]*V[3]);
  RM[1][2] = -2*(V[0]*V[2]+V[1]*V[3]);
  RM[2][2] = -V[0]*V[0]-V[1]*V[1]+V[2]*V[2]+V[3]*V[3];
  RM[3][2] =  0.;

  RM[0][3] =  0.;
  RM[1][3] =  0.;
  RM[2][3] =  0.;
  RM[3][3] =  1.;

 /*  print_matrice4x4(RM, "zuker RM", stderr);*/

  /* Solution eprouvee ! */
  try.x = - bc2.x; try.y = - bc2.y; try.z = - bc2.z;
  MkTrnsIIMat4x4(TY, try);
  MkTrnsIIMat4x4(TX, bc1);
  mulMat4x4(TY,RM,TMP);
  mulMat4x4(TMP, TX, M);

/*Ecrire ma matrice*/
 /*  {
    int i, j;
    FILE *pFiOut=stderr;
    fprintf(pFiOut, "zuker_superpose:: MATRIX M\n");
    for(i=0; i<4; i++){
      fprintf(pFiOut, "                 ");
      for(j=0; j<4; j++)
	fprintf(pFiOut, "%7.2f ", M[i][j]);
      fprintf(pFiOut, "\n");
    }
    fprintf(pFiOut, "\n");
  }
*/
   /* Now superpose the coordinates and Compute squared RMSd */
  for (aDot=0;aDot<len;aDot++) {
    p2=crdRotate(c2[aDot], M);
/*     fprintf(stderr, "%f %f %f -> %f\n",
	      p2.x,
	      p2.y,
	     p2.z,
	     SQUARED_DISTANCE(c1[aDot],p2));
 */
     squared_rms += SQUARED_DISTANCE(c1[aDot],p2);
  }
/*  fprintf(stderr, "len : %d sum %f RMS %f\n",len, squared_rms, squared_rms / (double) len);*/

  return squared_rms / (double) len;
}

/**************************************************/
/*AJOUT MATH*/
/* Calcul du RMS sans superposition optimale mais avec 
utilisation de la matrice prcdente
 * c1, c2 two sets of coordinates to superpose.
 * len : the number of coordiantes of c1, c2.
 *
 * M, the 4x4 transforamtion matrix to pass from c2 to c2 superposed onto c1.
 *
 * Input: 
 *   c1, c2, and len must be given. 
 *   M  a transformation matrix muts be given.
 * 
 * Output:
 *   c2 is superposed on c1.
 *
 * return value: 
 *   squared_rms (double) 
*/
/**************************************************/
double compute_RMS(COORDS *c1, COORDS *c2, int len, DtMatrix4x4 M)
{

  double squared_rms = 0.0L;
  int aDot;

/*Ecrire ma matrice*/
 /* {
    int ii, jj;
      FILE *pFiOut=stderr;
    fprintf(pFiOut, "compute_RMS:: MATRIX M\n");
    for(ii=0; ii<4; ii++){
      fprintf(pFiOut, "                 ");
      for(jj=0; jj<4; jj++)
	fprintf(pFiOut, "%7.2f ", M[ii][jj]);
      fprintf(pFiOut, "\n");
    }
    fprintf(pFiOut, "\n");
  }*/

  for (aDot=0;aDot<len;aDot++) 
    squared_rms += SQUARED_DISTANCE(c1[aDot],crdRotate(c2[aDot], M));
  

  return squared_rms / (double) len;
}

/********************************************/
/*         INTERNAL FUNCTIONS               */
/********************************************/

static double *random_vect(int n) {
  int i;

  double *v=NULL;

  NEWY(v,double, n);

  for (i=0; i<n; i++) 
    v[i]= (double)rand()/(RAND_MAX+1.0);
  return(v);
}

static double inner(double *x, double *y, int n) {
  int i;
  double sum;

  for (sum=0, i=0; i<n; sum+=x[i]*y[i],i++);
  return sum;
    
}

static double *product(double **A, double *x, int n) {
  int i, j;
  double sum;
  double *y=NULL;

  NEWY(y,double, n);

  for (i=0; i<n; i++) {
    sum=0;
    for (j=0; j<n; j++)
      sum+=A[i][j]*x[j];
    y[i]=sum;
  }
  return y;
}

static int lu_c (double a[4][4],  int n)
{
 int i,j,k,err;
 double pivot,coef;

 err=1;
 k=0;
 while (err==1 && k<n) {
  pivot=a[k][k];
  if(fabs(pivot)>=EPS) {
    for(i=k+1;i<n;i++) {
      coef=a[i][k]/pivot;
      for(j=k;j<n;j++)
	a[i][j] -= coef*a[k][j];
      a[i][k]=coef;
    }
  }
  else err=0;
  k++;
 }
 if(a[n-1][n-1]==0) err=0;
 return err;
}


static void resol_lu(double a[4][4], double *b, int n)
{
 int i,j;
 double sum;
 double y[n];
 y[0]=b[0];
 for(i=1;i<n;i++) {
  sum=b[i];
  for(j=0;j<i;j++)
    sum-=a[i][j]*y[j];
  y[i]=sum;
 }
 b[n-1]=y[n-1]/a[n-1][n-1];
 for(i=n-1;i>=0;i--) {
  sum=y[i];
  for(j=i+1;j<n;j++)
     sum-=a[i][j]*b[j];
  b[i]=sum/a[i][i];
 }
}

static int power(double *a[], int n, int maxiter, double eps, double *v, double *w) {
  int niter,i;
  double *y;
  double sum, l, normy, d;
  y=random_vect(n);
  niter=0;
  do {
    normy=sqrt(inner(y,y,n));
    for (i=0; i<n; i++) w[i]=y[i]/normy;
    y=product(a, w, n);
    l=inner(w,y,n);
    niter++;
    for (sum=0,i=0; i<n; i++) {
      d=y[i]-l*w[i];
      sum+=d*d;
    }
    d=sqrt(sum);
  } while (d>eps*fabs(l) && niter<maxiter);
  MYFREE(y);
  *v=l;
  return niter;
}

static double best_shift(double *a[], int n) {
  double m, M, s;
  double t, sum;
  int i, j;
  t=a[0][0];
  for (i=1; i<n; i++) t=max(t, a[i][i]);
  M=t;
  t=a[0][0];
  for (i=0; i<n; i++) {
    for (sum=0,j=0; j<n; j++)
      if (j!=i) sum+=fabs(a[i][j]);
    t=min(t, a[i][i]-sum);
  }
  m=t;
  s=-0.5*(M+m);
  for (i=0; i<n; i++) 
    a[i][i]=a[i][i]+s;
  return s;
}

static double lmax_estim (double a[4][4], int n) {
  double t, sum;
  int i, j;
  t=a[0][0];
  for (i=0; i<n; i++) {
    for (sum=0,j=0; j<n; j++)
      if (j!=i) sum+=fabs(a[i][j]);
    t=max(t, a[i][i]+sum);
  }
  return t;
}


static int shift_power(double *a, int n, int maxiter, double eps, double *v, double *w) {  
  double **tmp;
  double sh;
  int niter;
  int i,j;

  /* fprintf(stderr,"ESSAI avec shift_power.\n"); */

  tmp=alloc_mat(n, n);    
  /* copyMat(a, tmp, n, n); */
  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) {
      tmp[i][j] = a[i*n+j];
    }
  }
  sh=best_shift(tmp, n);   

  niter=power(tmp, n, maxiter, eps, v, w);
  *v=*v-sh;
  free_mat(tmp, n);
  return niter;
}

static int inverse_power(double a[4][4], int n, int maxiter, double eps, double *v, double *w) {
  int niter,i;
  double *y;
  double r, sum, l, normy, d;
  y=random_vect(n);
  niter=0;
  
  r=lmax_estim(a, n);
  for (i=0; i<n; i++) a[i][i]=a[i][i]-r;
  if (lu_c(a, n)==0) {
    /* fprintf(stderr,"ATTENTION ! cas singulier de inverse_power.\n"); */
    MYFREE(y);
    //exit(0);
    return 0;
  }

  do {
    normy=sqrt(inner(y,y,n));
    for (i=0; i<n; i++) {
      w[i]=y[i]/normy;
      y[i]=w[i];
    }
    resol_lu(a, y, n);
    l=inner(w,y,n);
    niter++;
    for (sum=0,i=0; i<n; i++) {
      d=y[i]-l*w[i];
      sum+=d*d;
    }
    d=sqrt(sum);
  } while (d>eps*fabs(l) && niter<maxiter);
  MYFREE(y);
  *v=r+1.0/l;
  return niter;
}

static int largestEV4(double R[4][4], double v[4], double *vp)
{
/*  double iw[4];
  double dx,dy,dz, dt;*/
/*  int aCycle;*/
  double M2[4][4];
  int rs;

  memcpy(M2,R,sizeof(double)*16);

  rs = inverse_power(R, 4, 10000, 1.e-8, vp, v);

  if (!rs) 
    return shift_power(&M2[0][0], 4, 10000, 1.e-8, vp, v);

  return rs;

}


/* ==================================================================
 * Calcule la matrice d'inertie des points
 *
 * id      : indices des atomes dans At
 * from,tto: indices extremes
 * P       : le barycentre
 *
 * Attention XMean et YMin sont des pointeurs car ils sont en entre et sortie
 * ==================================================================
 */
static DtMatrix3x3 *XYCov(DtMatrix3x3 *pM, 
		   COORDS *X, 
		   COORDS *Y, 
		   COORDS *Xmean, 
		   COORDS *Ymean, 
		   int aSze)
{
  int i;
  double Xx,Xy,Xz;
  double Yx,Yy,Yz;
  double daSze;

/*   fprintf(stdout,"inertia, len %d\n",aSze); */

  /* X average*/
  Xmean->x = Xmean->y = Xmean->z = 0.;
  for (i=0;i<aSze;i++) {
    Xmean->x += X[i].x;
    Xmean->y += X[i].y;
    Xmean->z += X[i].z;
  }
  if (aSze) {
    daSze = (double) aSze;
    Xmean->x /= daSze;
    Xmean->y /= daSze;
    Xmean->z /= daSze;
  }

  /* Y average*/
  Ymean->x = Ymean->y = Ymean->z = 0.;
  for (i=0;i<aSze;i++) {
    Ymean->x += Y[i].x;
    Ymean->y += Y[i].y;
    Ymean->z += Y[i].z;
  }
  if (aSze) {
    daSze = (double) aSze;
    Ymean->x /= daSze;
    Ymean->y /= daSze;
    Ymean->z /= daSze;
  }

  /* Covariance matrix */
  if (pM == NULL) {
    CALLOCY(pM, DtMatrix3x3, 1);
  } else {
    memset((void *) pM, 0, sizeof(DtMatrix3x3));
  }

  for (i=0;i<aSze;i++) {
    Xx = (double) X[i].x - Xmean->x;
    Xy = (double) X[i].y - Xmean->y;
    Xz = (double) X[i].z - Xmean->z;
    Yx = (double) Y[i].x - Ymean->x;
    Yy = (double) Y[i].y - Ymean->y;
    Yz = (double) Y[i].z - Ymean->z;

    (*pM)[0][0] += Xx*Yx;
    (*pM)[0][1] += Xx*Yy;
    (*pM)[0][2] += Xx*Yz;

    (*pM)[1][0] += Xy*Yx;
    (*pM)[1][1] += Xy*Yy;
    (*pM)[1][2] += Xy*Yz;

    (*pM)[2][0] += Xz*Yx;
    (*pM)[2][1] += Xz*Yy;
    (*pM)[2][2] += Xz*Yz;
  }

/*  {
    int ii;
    fprintf(stderr, "XYCov::\n");
    for(ii=0; ii<3; ii++)
      fprintf(stderr, "%f %f %f\n", 
	      (*pM)[ii][0],
	      (*pM)[ii][1],
	      (*pM)[ii][2]);
    
  }*/
  return pM;
}

/* ------------------------------------------------------------------ 
    Matrice 4x4 Translation
   ---------- PREMULTIPLIE -> Y = XM (X vecteur ligne) ----------- 
   ------------------------------------------------------------------ */
static void MkTrnsIIMat4x4(DtMatrix4x4 m, COORDS tr)
{
 m[0][0] = 1.;    m[0][1] = 0.;    m[0][2] = 0.;    m[0][3] = 0.;
 m[1][0] = 0.;    m[1][1] = 1.;    m[1][2] = 0.;    m[1][3] = 0.;
 m[2][0] = 0.;    m[2][1] = 0.;    m[2][2] = 1.;    m[2][3] = 0.;
 m[3][0] = tr.x; m[3][1] = tr.y; m[3][2] = tr.z; m[3][3] = 1.;
}

/* Matrices 4 x 4 (a x b dans c) ------------------------------------ */
static void mulMat4x4(DtMatrix4x4 a,DtMatrix4x4 b,DtMatrix4x4 c)
{
  int i,j,k;
  boucle(i,0,4) {
    boucle(j,0,4) {
      c[i][j] = 0.;
      boucle(k,0,4) c[i][j] += a[i][k]*b[k][j];
    }
  }
}



/**************************************************/
static double **alloc_mat(int n, int m) {
  int i;
  double **mat = NULL;

  CALLOCY(mat, double*, m);
  for (i=0; i<n; i++)
    CALLOCY(mat[i], double, m);
  
  return mat;
}

/**************************************************/
static void free_mat(double **mat, int n){
  int i;
  for (i=0; i<n; i++)
    MYFREE(mat[i]);
  MYFREE(mat);
}
