/* -------------------------------------------------------------------- */
/* @file: libphipsi.c							*/
/*									*/
/* Calcul des angles et liaisons a l'oeuvre dans les proteines		*/
/*									*/
/* calcul des angles phi, psi, omega 					*/
/* phi = (C-1,N,CA,C)							*/
/* psi = (N,CA,C,N+1)							*/
/* omega = (CA,C,N+1,CA+1)						*/
/*									*/
/* calcul de chi1, chi2, chi3 d'une entree de la pdb			*/
/*									*/
/* calcul de alpha, tau et r12						*/
/* (Levitt J. Mol. Biol. 1976 ,vol. 104, 59-107)			*/
/*									*/
/* ici les angles sont defini comme suit : 				*/
/* alpha = (CA-1,CA,CA+1,CA+2)						*/
/* tau = (CA_CA-1,CA_CA+1)						*/
/*									*/
/* calcul de la distance (r13) entre le CA+1 et le CA-1 pour un residu	*/
/*									*/
/* les angles sont donnes entre [-180.0 , 180.0 [      			*/
/* si un angle ne peut etre determine pour une raison			*/
/* ou une autre dans un residu, il prend la valeur			*/
/* 500 (NO_ANGLE) de meme pour la distance				*/
/*									*/
/* -------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <math.h>

#include "libphipsi.h"

/* -------------------------------------------------------------------- */
/* local vars								*/
/* -------------------------------------------------------------------- */

static char *sResiduNames        = "/UNK/ALA/CYS/ASP/GLU/PHE/GLY/HIS/ILE/LYS/LEU/MET/ASN/PRO/GLN/ARG/SER/THR/VAL/TRP/TYR/MSE/CSE/ASX/GLX/CEA/HYP/ACD/HSE/HYL/ALB/ARO/PCA/BAS/SAR/BET/TAU/";
static char *sResiduNoChi1       = "/GLY/ALA/";
static char *sResiduNoChi2       = "/CYS/SER/THR/VAL/";
static char *sResiduNoChi3       = "/ILE/LEU/ASP/ASN/";

/* -------------------------------------------------------------------- */
/* macros								*/
/* -------------------------------------------------------------------- */

#ifndef NEW
#define NEW(object)		(object*)malloc(sizeof(object)) 
#define NEWN(object, dim) 	(object*)malloc((unsigned)(dim) * sizeof(object))
#define FREE(object)		free(object)
#endif

#ifndef MYFREE
#define MYFREE(x)	if ((x)) free(x)
#endif

#ifndef REALLOC
#define REALLOC(ptr, typ, dim)  {ptr=(typ*)realloc((void *) (ptr), (unsigned long)(dim) * sizeof(typ));\
                                if((ptr)==NULL){\
                                  fprintf(stderr,"pb realloc avec %s (%d %s)\n",\
                                  #ptr,__LINE__,__FILE__);exit(99);}}

#endif

#ifndef NEWY
#define NEWY(var, typ, dim)  {var=(typ*)malloc((size_t)(dim)*sizeof(typ));\
                              if((var)==NULL){\
                                  fprintf(stderr,"pb malloc avec %s (%d %s)\n",\
                                  #var,__LINE__,__FILE__);exit(99);}}
#endif

#define GET_DIST(O1, O2) (float)(sqrt((double)((O1->x - O2->x)*(O1->x - O2->x) + \
				(O1->y - O2->y)*(O1->y - O2->y) + \
				 (O1->z - O2->z)*(O1->z - O2->z))))


# define LINE_LENGTH	128	/* length max of a line in pdb file	*/


#define PI		3.14159L



#define HAS_NO_CA_ATOM		'A'
#define HAS_NO_CB_ATOM		'B'
#define HAS_NO_N_ATOM		'N'
#define HAS_NO_CO_ATOM		'C'
#define HAS_NO_CG_ATOM		'G'
#define HAS_NO_CD_ATOM		'D'
#define HAS_NO_CE_ATOM		'E'
#define AT_START_OF_CHAIN	'S'
#define AT_END_OF_CHAIN		'F'


#define D_PRINTCHAIN 55

/* -------------------------------------------------------------------- */
/* prototypes   internes       						*/
/* -------------------------------------------------------------------- */
static	ResAtom	*resatom_cpy(ResAtom *dst, ResAtom *src);
static	ResAtom	*atom_resatom_cpy(ResAtom *dst, Atom *src);
static	ResAtom	*get_named_atom(char *name, Residu *r, ResAtom *dst);
static	double  resatom_distca( ResAtom *at);
static	void	angle_error(char *nam1, int num1, char *spos,
		   char *nam2, int num2, char smiss, int subroutine);
#if 0
static	char	*low(char *str);
#endif
static	char	*trimr(char *str, int n);
static	Residu	*new_residu(void);
static	void	free_residu(Residu *r);
#ifdef D_PRINTRES
static	void	PrintRes(Residu *r);
#endif
static	ResAtom	*put_atom_in_res(Atom *atom, Residu *res);
static	double	atom_dihed(ResAtom *ra);
static	double	atom_angle(ResAtom *ra);
static	int	compute_phipsi(Residu *r, ResAtom *ra,
			int do_phi, int do_psi, int do_ome);
static	int	compute_chis(Residu  *r, ResAtom *ra,
			int do_chi1, int do_chi2, int do_chi3);
static	int	compute_alpha_tau(Residu  *r, ResAtom *ra,
		int do_tau, int do_alpha);
static	char	*pdb_field(char *line, int start, int len);
static	void	myInfoOnStderr(char *info);


#ifdef D_PRINTRES
/* -------------------------------------------------------------------- */
/* Print a residue							*/
/* -------------------------------------------------------------------- */
static	void	PrintRes(Residu *r)
{
	fprintf(stderr, "%3s %d %d\n", r->name, r->num, r->pdbnum) ;
}
#endif
#ifdef D_PRINTCHAIN
/* -------------------------------------------------------------------- */
/* Print all chains							*/
/* -------------------------------------------------------------------- */
static	void	PrintChain(tyChain *tChain, int nbChain)
{
/*   Residu *pChain; */
/*   int nbRes; */

  int i;
  Residu *r;

    fprintf(stderr, "%d chain\n",  nbChain);

  for(i=0; i<nbChain; i++){
    r=tChain[i].pChain;
    fprintf(stderr, "Chain %d %c first res %d%c nbRes %d\n",  i, r->chain, r->pdbnum, r->inser, tChain[i].nbRes);
  }

}
#endif

#if 0	/* old	*/ 
/* Dans libabiutils*/
/* -------------------------------------------------------------------- */
/* gestion (simplifiee) des erreurs					*/
/* imprime un message et arrete le massacre				*/
/* -------------------------------------------------------------------- */
void MyError(FILE *stream, int stat, char *fmt, ...)
{
   va_list args;

   va_start(args,fmt);
   (void)vfprintf(stream, fmt, args);
   (void)fputc((int)'\n', stream);
   va_end(args);
   (void)fflush(stream);

   if (stat > 0)
      exit(stat);

   return;

}

void myerror(char *msg, int status)
{
	if (status) {
	  fprintf(stderr, " *error* %s\n", msg);
	  exit(status);
	}
	else
	  fprintf(stderr, " *warning*%s\n", msg);

	return;
}
#endif

/* -------------------------------------------------------------------- */
/* gestion (simplifiee) des informations (pas vraiment erreurs)		*/
/* imprime un message sur stderr					*/
/* -------------------------------------------------------------------- */
static void myInfoOnStderr(char *info)
{
	fprintf(stderr,"%s\n", info);
}

/* -------------------------------------------------------------------- */
/* utility : lower case a string					*/
/* -------------------------------------------------------------------- */
#if 0
static char *low(char *str)
{
	char *s;

	for (s = str ; *s ; s++)
	   if (isupper(*s))
		*s = tolower(*s);

	return str;
}
#endif
/* -------------------------------------------------------------------- */
/* utility : trim spaces on the right of string				*/
/*	     (starting from pos n)					*/
/* -------------------------------------------------------------------- */
static char *trimr(char *str, int n)
{
	char *s;

	s = str + n;

	while ((s > str) && isspace((int)(*s)))
		*s-- = '\000';

	return str;
}


/* -------------------------------------------------------------------- */
/* creation d'un residu							*/
/* -------------------------------------------------------------------- */
static Residu *new_residu(void)
{
	Residu *r;

	if (! (r = NEW(Residu))) {
	   MyError(stderr, 0, "new_residu:: not enough memory");
	   return NULL;
	}

	memset(r, 0, sizeof(Residu));

	r->nb_neighbours = 0;

	return r;
}

/* -------------------------------------------------------------------- */
/* liberation d'une liste de residus				                	*/
/* -------------------------------------------------------------------- */
void free_list_residus(Residu *first_res)
{
    Residu *cur_res, *prev_res;

    prev_res = first_res;
    cur_res = first_res->next;

    while (prev_res != NULL) {
      free_residu(prev_res);
      prev_res = cur_res;
      if (cur_res) {
	cur_res = cur_res->next;
      }
    }
    
    return;
}

void free_list_atom (ResAtom *ra){

    ResAtom *cur_ra, *prev_ra;

    if (! ra)
            return;

    prev_ra = ra;
    cur_ra = ra->next;
    
    while (prev_ra != NULL) {
      free(prev_ra);
      prev_ra = cur_ra;
      if (cur_ra) {
	cur_ra = cur_ra->next;
      }
    }

    return;
}

/* -------------------------------------------------------------------- */
/* liberation d'un residu							*/
/* -------------------------------------------------------------------- */
static void free_residu(Residu *r)
{

	if (r) {
	   free_list_atom(r->first_atom);
	   free(r);
	   
	}
	return;
}

/* -------------------------------------------------------------------- */
/* copy resatom -> resatom						*/
/* -------------------------------------------------------------------- */
static ResAtom *resatom_cpy(ResAtom *dst, ResAtom *src)
{
	strcpy(dst->name, src->name);
	dst->x =  src->x;
	dst->y =  src->y;
	dst->z =  src->z;
	dst->tfactor =  src->tfactor;
	
	return dst;
}

/* -------------------------------------------------------------------- */
/* copy atom -> resatom							*/
/* -------------------------------------------------------------------- */
static ResAtom *atom_resatom_cpy(ResAtom *dst, Atom *src)
{
	strcpy(dst->name, src->name);
	dst->x =  src->x;
	dst->y =  src->y;
	dst->z =  src->z;
	dst->tfactor =  src->tfactor;
/* 	fprintf(stderr, "ATOM %s tFacor: %f\n", src->name, src->tfactor); */
	return dst;
}

/* -------------------------------------------------------------------- */
/* adds an atom in residue (returns the associated, new 		*/
/* resatom)								*/
/* returns the Resatom (or NULL if it can't be allocated )		*/ 
/* -------------------------------------------------------------------- */
static ResAtom *put_atom_in_res(Atom *atom, Residu *res)
{
	ResAtom *ra = NULL;

	if (! (ra = NEW(ResAtom)))
	
	   MyError(stderr, 0, "put_atom_in_res:: not enough memory");

	else {
	
	   atom_resatom_cpy(ra, atom);
	   ra->next = NULL;

	   if (! res->first_atom)
	      res->first_atom = ra;
	   else
	      res->last_atom->next = ra;

	   res->last_atom  = ra;
	}

	return ra;
}

/* -------------------------------------------------------------------- */
/* maintains the list of the neighbours of a residue			*/
/* returns the number of neighbours or -1 if too much neighbours	*/
/* -------------------------------------------------------------------- */
int set_residue_neighbours(Residu *res, Residu *first)
{	
	Residu	*r;

	res->nb_neighbours = 0;

	for (r = first; r; r = r->next) {
	   if (r == res)
	      continue;
	   /* d = GET_DIST(r, res); i*//* debug */
	   if (GET_DIST(r, res) <= DIST_MAX_BETWEEN_NEIGHBOUR_RESIDUE) {
	       if (++res->nb_neighbours >= MAX_NEIGHBOURS ) {
		   MyError(stderr, 4, "set_residue_neighbours:: trop de voisins");
		   return -1;
	       }
	       else
		   res->neighbour[res->nb_neighbours-1] = r;
	   }
	}
	
	return res->nb_neighbours;
}



/* -------------------------------------------------------------------- */
/* find the geometric center of atoms in a residue			*/
/* and set the x, y, z field of the Residu structure			*/
/* return the number of atom found					*/
/* -------------------------------------------------------------------- */
int      set_residue_geocenter(Residu *r)
{
	ResAtom *a;
	int	n = 0;
	float	sx = 0.0, sy = 0.0, sz = 0.0;

	for (a = r->first_atom; a != NULL; a = a->next) {
	   sx += a->x;
	   sy += a->y;
	   sz += a->z;
	   n++;
	}

	if (n > 0) {
	   r->x = sx/(float)n;
	   r->y = sy/(float)n;
	   r->z = sz/(float)n;
	}
	else
	   r->x = r->y = r->z = 0.;
	
	return n;
}

/* -------------------------------------------------------------------- */
/* get the maximum distance of an atom from the				*/
/* geometric center of the residue					*/
/* -------------------------------------------------------------------- */
float get_dist_max_from_geocenter(Residu *r)
{
	float d, dmax = -1.0;
	ResAtom	*a;

	for (a = r->first_atom; a != NULL; a = a->next) {
	   d = GET_DIST(r, a);
	   if (d > dmax) {
		dmax = d;
	   }
	}

	return dmax;
}

/* -------------------------------------------------------------------- */
/* get the x y z coordinates of CA of residue r				*/
/* -------------------------------------------------------------------- */
int     get_CA_xyz(Residu *r, float *x, float *y, float *z, int verbose) {
	ResAtom ra;

	if ( ! get_named_atom( "CA", r, &ra)) {
	   if (verbose)
		   fprintf(stderr, "get_CA_xyz:%3s %4d has no CA atom\n",
			   r->name, r->num);
	   *x = *y = *z = 0.0;
	}
	*x = ra.x;
	*y = ra.y;
	*z = ra.z;

	return 0;
}
/* -------------------------------------------------------------------- */
/* get the x y z coordinates of CA of residue r				*/
/* -------------------------------------------------------------------- */
float	get_CA_TFactor(Residu *r, int verbose)
{
	ResAtom ra;
	float t=0.0;

	if ( ! get_named_atom( "CA", r, &ra)) {
	   if (verbose)
		   fprintf(stderr, "get_CA_TFactor:%3s %4d has no CA atom\n",
			   r->name, r->num);
	   t= -1.0;
	}


	t = ra.tfactor;

	return(t);
}

/* -------------------------------------------------------------------- */
/* extract atom (by name) from residu					*/
/* if found, copy atom into dst and returns atom			*/
/* else, returns NULL							*/
/* -------------------------------------------------------------------- */
ResAtom *get_named_atom(char *name, Residu *r, ResAtom *dst)
{
	ResAtom *atm;

	for (atm = r->first_atom ; atm ; atm = atm->next)
	  if (! strcmp(name, atm->name)) {
		resatom_cpy(dst, atm);
		return atm;
	  }

	return NULL;

}


/*------------------------------------------------------------------------------*/
/* compute the distance between CA in Resatom at[0] and at[1]			*/
/*------------------------------------------------------------------------------*/
static double   resatom_distca( ResAtom *at)
{
	return( sqrt(   ((at+1)->x - at->x)*((at+1)->x - at->x) +
			((at+1)->y - at->y)*((at+1)->y - at->y) +
			((at+1)->z - at->z)*((at+1)->z - at->z)));
}

/*------------------------------------------------------------------------------*/
/* compute the distance between CA-1 and CA+1					*/
/*------------------------------------------------------------------------------*/
int  get_distca_1_3( Residu *r, int verbose)
{
	ResAtom at[2];
						/* respectively: at[0] -> CA-1	*/
						/*               at[1] -> CA+1	*/
	r->dis13 = 0.0;

#define ERRORDIST(X,NX,S,Y,NY,CAR)	{ r->dis13 = NO_ANGLE;			\
					if (verbose) angle_error(X,NX,S,Y,NY,CAR,4);		\
					return 1; }

	if ( (! r->prev) || (r->num != r->prev->num + 1))

	   ERRORDIST(r->name, r->num, NULL, NULL, 0, AT_START_OF_CHAIN)

						/* find CA atom of prev residue	*/
	else if ( ! get_named_atom( "CA", r->prev, at))

	   ERRORDIST(r->prev->name, r->prev->num, "preceding", r->name, r->num,
	   							HAS_NO_CA_ATOM)

	if (( ! r->next) || (r->num != r->next->num - 1) )

	   ERRORDIST(r->name, r->num, NULL, NULL, 0, AT_END_OF_CHAIN)

						/* find CA atom of next residue */
	else if ( ! get_named_atom( "CA", r->next, at+1))

	   ERRORDIST(r->next->name, r->next->num, "following", r->name, r->num,
	   							HAS_NO_CA_ATOM)

	r->dis13 = (float)resatom_distca(at);

	return 0;

#undef ERRORDIST

}


/*------------------------------------------------------------------------------*/
/* compute the distance between CA and CA+1					*/
/*------------------------------------------------------------------------------*/
int get_distca_1_2( Residu *r, int verbose)
{
	ResAtom at[2];
						/* respectively: at[0] -> CA	*/
						/*               at[1] -> CA+1	*/

#define ERRORDIST(X,NX,S,Y,NY,CAR)		{ r->dis12 = NO_ANGLE;		\
						if (verbose) angle_error(X,NX,S,Y,NY,CAR,3);	\
						return 1; }
	r->dis12 = 0.0;

	if (! r) {
	   MyError(stderr, 0, "get_distca_1_2:: residue does not exist");
	   return 1;
	}
					/* find CA atom of current residue */
	else if ( ! get_named_atom( "CA", r, at))
	   ERRORDIST(r->name, r->num, NULL, NULL, 0, HAS_NO_CA_ATOM)

	if (( ! r->next) || (r->num != r->next->num - 1) )
	   ERRORDIST(r->name, r->num, NULL, NULL, 0, AT_END_OF_CHAIN)
					/* find CA atom of next residue */
	else if ( ! get_named_atom( "CA", r->next, at+1))
	   ERRORDIST(r->next->name, r->next->num, "following", r->name, r->num,
	   						HAS_NO_CA_ATOM)

	r->dis12 = (float)resatom_distca(at);
    if (r->dis12 >DIST_MAX_BETWEEN_NEIGHBOUR_RESIDUE){
        if (verbose) fprintf(stderr,"erreur distance : trou\n");
      return 1;}

	return 0;

#undef ERRORDIST
}

/* -------------------------------------------------------------------- */
/* put a warning message when an angle cannot be			*/
/* computed (somewhat verbose ?)					*/
/* -------------------------------------------------------------------- */
static void angle_error(char *nam1, int num1, char *spos,
		char *nam2, int num2, char smiss, int subroutine)
{
	char subroutine_name[20], smissmsg[50], strerr[256];

#define SUBCPY(X)	strcpy(subroutine_name, X); break

	switch(subroutine) {

	   case 0: SUBCPY("get_phi_psi   ");

	   case 1: SUBCPY("get_chi       ");

	   case 2: SUBCPY("get_alpha_tau ");

	   case 3: SUBCPY("get_distca_1_2");

	   case 4: SUBCPY("get_distca_1_3");

	}

#undef SUBCPY

#define SMISSCPY(X)	strcpy(smissmsg, X); break

	switch(smiss) {

	   case HAS_NO_CA_ATOM: SMISSCPY("has no CA atom");

	   case HAS_NO_CB_ATOM:SMISSCPY("has no CB atom");

	   case HAS_NO_N_ATOM: SMISSCPY("has no N atom");

	   case HAS_NO_CO_ATOM: SMISSCPY("has no CO atom");

	   case HAS_NO_CG_ATOM: SMISSCPY("has no CG(1), OG(1) or SG atom");

	   case HAS_NO_CD_ATOM: SMISSCPY("has no CD(1), OD(1), SD or ND(1) atom");

	   case HAS_NO_CE_ATOM: SMISSCPY("has no CE, CE(1), OE(1), NE or NE(1) atom");

	   case AT_START_OF_CHAIN: SMISSCPY("at start of chain");

	   case AT_END_OF_CHAIN: SMISSCPY("at end of chain");

	}

#undef SMISSCPY

	if (nam2) {

	   sprintf(strerr, "%s :: residue %s %4d %s residue %s %4d %s",
		subroutine_name, nam1, num1, spos,nam2, num2, smissmsg);

	   myInfoOnStderr(strerr);

	}

	else {

	   sprintf(strerr, "%s :: residue %s %4d %s",
		subroutine_name, nam1, num1, smissmsg);

	   myInfoOnStderr(strerr);

	}

}

/* -------------------------------------------------------------------- */
/* computes the angle defined by the 3 points				*/
/* given in ra. returns angle in [-180, 180[ deg.			*/
/* -------------------------------------------------------------------- */
static double atom_angle(ResAtom *ra)
{
	double	xa, ya, za, xb, yb, zb,
		xc, yc, zc, np1np2, ang;

				/* calcul des composantes vectorielles	*/

	xa = ra->x - (ra+1)->x;
	ya = ra->y - (ra+1)->y;
	za = ra->z - (ra+1)->z;

	xb = (ra+2)->x - (ra+1)->x;
	yb = (ra+2)->y - (ra+1)->y;
	zb = (ra+2)->z - (ra+1)->z;


				/* (produit des normes)			*/

	np1np2 = sqrt((xa*xa+ya*ya+za*za) * (xb*xb+yb*yb+zb*zb));

	if (np1np2 == 0.0L) 
		return NO_ANGLE;

				/* produit vectoriel 1*2		*/

	xc =   ya*zb - yb*za;
	yc = - xa*zb + xb*za;
	zc =   xa*yb - xb*ya;
				/* arcsinus de norme du produit vect.	*/

	ang = asin(sqrt(xc*xc+yc*yc+zc*zc)/np1np2);

				/* ang is between -Pi/2 and Pi/2	*/
				/* we need to know the sign		*/
				/* of cosinus (produit scalaire)	*/
	
	if ((xa*xb+ya*yb+za*zb) <  0)			/* cos is < 0	*/
		ang = PI - ang;

        ang = ang * 180.0L / PI;			/* rad -> deg	*/
	if (ang >= 180.0L)			 /* [-180.0,180.0[ range */
	  ang = -180.0L;


	return ang;

}

/* -------------------------------------------------------------------- */
/* computes the dihedral angle defined by the 4 points			*/
/* given in ra. returns angle in [-180, 180[ deg.			*/
/* -------------------------------------------------------------------- */
static double atom_dihed(ResAtom *ra)
{
	double	xa, ya, za, px1, py1, pz1,
		xb, yb, zb, px2, py2, pz2,
		xc, yc, zc, px3, py3, pz3,
		np1np2, dihed, tmp;

				/* calcul des composantes vectorielles	*/

	xa = (ra+1)->x - ra->x;
	ya = (ra+1)->y - ra->y;
	za = (ra+1)->z - ra->z;

	xb = (ra+2)->x - (ra+1)->x;
	yb = (ra+2)->y - (ra+1)->y;
	zb = (ra+2)->z - (ra+1)->z;

	xc = (ra+3)->x - (ra+2)->x;
	yc = (ra+3)->y - (ra+2)->y;
	zc = (ra+3)->z - (ra+2)->z;

					/* 1er produit vectoriel	*/

	px1 =   ya*zb - yb*za;
	py1 = - xa*zb + xb*za;
	pz1 =   xa*yb - xb*ya;

					/* 2eme produit vectoriel	*/

	px2 =   yb*zc - yc*zb;
	py2 = - xb*zc + xc*zb;
	pz2 =   xb*yc - xc*yb;

				/* (produit des normes)			*/

	np1np2 = sqrt((px1*px1+py1*py1+pz1*pz1) * (px2*px2+py2*py2+pz2*pz2));

	if (np1np2 == 0) 
		return NO_ANGLE;

				/* arccosinus du produit scalaire	*/
				/* des produits vectoriels 		*/
/* printf("angle 0000: %e %e \n", px1*px2+py1*py2+pz1*pz2, (px1*px2+py1*py2+pz1*pz2)/np1np2);

 printf("angle 8888: %e\n",acos((px1*px2+py1*py2+pz1*pz2)/np1np2));*/


       tmp = (px1*px2+py1*py2+pz1*pz2)/np1np2; 
        if(tmp > 1.0)
            tmp = 1.0;
        else if(tmp < -1.0)
             tmp = -1.0;
             
	dihed = acos(tmp);

				/* dihed is between 0 and Pi		*/
				/* we need the sign			*/
				/* produit vectoriel des produits vect.	*/

	px3 =   py1*pz2 - py2*pz1;
	py3 = - px1*pz2 + px2*pz1;
	pz3 =   px1*py2 - px2*py1;
	
				/* produit scalaire du 2eme vect	*/
				/* avec prod. vect. des prod. vect.	*/
	
	if ((px3*xb + py3*yb + pz3*zb) <  0)		/* sine is < 0	*/
		dihed = -dihed;

        dihed = dihed * 180.0L / PI;			/* rad -> deg	*/
	if (dihed >= 180.0L)			/* [-180.0,180.0[ range	*/
	  dihed = -180.0L;
	return dihed;

}

/* -------------------------------------------------------------------- */
/* computes phi, psi and omega dihed. angles				*/
/* put result in r and returns 1 if all angles defined			*/
/* atoms coordinates are given in ra (see below)			*/
/* -------------------------------------------------------------------- */
static int compute_phipsi(Residu *r, ResAtom *ra,
		int do_phi, int do_psi, int do_ome)
{
	r->phi = (do_phi ? atom_dihed(ra)     : NO_ANGLE);
	r->psi = (do_psi ? atom_dihed(ra + 1) : NO_ANGLE);
	r->ome = (do_ome ? atom_dihed(ra + 2) : NO_ANGLE);

	return (    (r->phi != NO_ANGLE)
		 && (r->psi != NO_ANGLE)
		 && (r->ome != NO_ANGLE));
}

/* -------------------------------------------------------------------- */
/* computes chi1, chi2 and chi3 dihed. angles				*/
/* put result in r and returns 1 if all angles defined			*/	
/* atoms coordinates are given in ra (see below)			*/	
/* -------------------------------------------------------------------- */
static int compute_chis(Residu  *r, ResAtom *ra,
		int do_chi1, int do_chi2, int do_chi3)
{
	r->x1 = (do_chi1 ? atom_dihed(ra)     : NO_ANGLE);
	r->x2 = (do_chi2 ? atom_dihed(ra + 1) : NO_ANGLE);
	r->x3 = (do_chi3 ? atom_dihed(ra + 2) : NO_ANGLE);

	return (    (r->x1 != NO_ANGLE)
		 && (r->x2 != NO_ANGLE)
		 && (r->x3 != NO_ANGLE));
}

/* -------------------------------------------------------------------- */
/* check  phi, psi and omega computability, compute 			*/
/* them (put result in r) and return 1 if all angles			*/
/* are defined								*/
/* -------------------------------------------------------------------- */

#define CHECK_ANGLES 	if (! (dphi || dpsi || dome)) \
			   return compute_phipsi(r, ra, 0, 0, 0)

int get_phi_psi(Residu *r, int verbose)
{
	int	dphi, dpsi, dome;	/* computability flags		*/
	ResAtom ra[6];			/* ra[0] -> CO-1		*/
					/* ra[1] -> NA  		*/
					/* ra[2] -> CA  		*/
					/* ra[3] -> CO  		*/
					/* ra[4] -> NA+1		*/
					/* ra[5] -> CA+1		*/

	if (! r) {
	   MyError(stderr, 0, "get_phi_psi:: residue does not exist");
	   return 0;
	}

	dphi = dpsi = dome = 1;		/* assume all is ok 		*/

	/* ------------------------------------------------------------ */
	/* first of all, we need CA and CO atoms of the current residue	*/
	/* for phi, psi and omega					*/
	/* ------------------------------------------------------------ */

	if (! get_named_atom("CA", r, ra + 2)) {
	   if (verbose)
		   angle_error(r->name, r->num, NULL, NULL, 0, HAS_NO_CA_ATOM, 0);
	   dphi = dpsi = dome = 0;
	}

	CHECK_ANGLES;

	if (! get_named_atom("C", r, ra + 3)) {
	   if (verbose)
		   angle_error(r->name, r->num, NULL, NULL, 0,HAS_NO_CO_ATOM, 0);
	   dphi = dpsi = dome = 0;
	}

	CHECK_ANGLES;

	/* ------------------------------------------------------------ */
	/* we need the N atom too (for phi and psi but not for omega)	*/
	/* ------------------------------------------------------------ */

	if (! get_named_atom("N", r, ra + 1)) {
	   if (verbose)
		   angle_error(r->name, r->num, NULL, NULL, 0,HAS_NO_N_ATOM, 0);
	   dphi = dpsi = 0;
	}

	CHECK_ANGLES;

	/* ------------------------------------------------------------ */
	/* phi angle is undefined for the first residue of a chain	*/
	/* else, we need the CO atom of previous residue for phi	*/
	/* ------------------------------------------------------------ */
	
	if (! (r->prev && (r->num == r->prev->num + 1))) {
	   if (verbose)
		   angle_error(r->name, r->num, NULL, NULL, 0, AT_START_OF_CHAIN, 0);
	   dphi = 0;
	}
	else if (! get_named_atom("C", r->prev, ra + 0)) {
	   if (verbose)
		   angle_error(r->prev->name, r->prev->num, "preceding", 
		       r->name, r->num, HAS_NO_CO_ATOM, 0);
	   dphi = 0;
	}

	CHECK_ANGLES;

	/* ------------------------------------------------------------ */
	/* psi and omega are undefined for the last residue of a chain	*/
	/* else, we need the NA and CA atoms in next residue to compute	*/
	/* them								*/
	/* ------------------------------------------------------------ */
	
	if (! (r->next && (r->next->num == r->num + 1))) {
	   if (verbose)
		   angle_error(r->name, r->num, NULL, NULL, 0, AT_END_OF_CHAIN, 0);
	   dpsi = dome = 0;
	}
	else if (! get_named_atom("N", r->next, ra + 4)) {
	   if (verbose)
		   angle_error(r->next->name, r->next->num, "following", 
		   r->name, r->num, HAS_NO_N_ATOM, 0);
	   dpsi = dome = 0;
	}
	else if (! get_named_atom("CA", r->next, ra + 5)) {
	   if (verbose)
		   angle_error(r->next->name, r->next->num, "following", 
		   r->name, r->num, HAS_NO_CA_ATOM, 0);
	   dome = 0;
	}

	CHECK_ANGLES;

	return compute_phipsi(r, ra, dphi, dpsi, dome);
}

#undef CHECK_ANGLES

/* -------------------------------------------------------------------- */
/* check  chi1, 2, 3 computability, compute them			*/
/* put result in r) and return 1 if all angles are			*/
/* defined								*/
/* -------------------------------------------------------------------- */

#define CHECK_ANGLES 	if (! (dchi1 || dchi2 || dchi3)) \
			   return compute_chis(r, ra, 0, 0, 0)

int get_chi(Residu *r, int verbose)
{
	int	dchi1, dchi2, dchi3;	/* computability flags		*/	
	ResAtom ra[6];			/* ra[0] -> NA			*/
					/* ra[1] -> CA  		*/
					/* ra[2] -> CB  		*/
					/* ra[3] -> CG  		*/
					/* ra[4] -> NA+1		*/
					/* ra[5] -> CA+1		*/

	if (! r) {
	   MyError(stderr, 0, "get_chi:: residue does not exist");
	   return 0;
	}

	dchi1 = dchi2 = dchi3 = 1;	/* assume all is ok		*/

	/* ------------------------------------------------------------ */
	/* first, check residus for which chi's cannot be computed	*/
	/* (silent)							*/
	/* ------------------------------------------------------------ */

	if (strstr(sResiduNoChi1, r->name))
	   dchi1 = dchi2 = dchi3 = 0;
	else if (strstr(sResiduNoChi2, r->name))
	   dchi2 = dchi3 = 0;
	else if (strstr(sResiduNoChi3, r->name))
	   dchi3 = 0;

	CHECK_ANGLES;

	/* ------------------------------------------------------------ */
	/* we need NA atom of current residue for chi1			*/
	/* ------------------------------------------------------------ */

	if (dchi1 && (! get_named_atom("N", r, ra))) {
	   if (verbose)
		   angle_error(r->name, r->num, NULL, NULL, 0, HAS_NO_N_ATOM, 1);
	   dchi1 = 0;
	}

	CHECK_ANGLES;

	/* ------------------------------------------------------------ */
	/* we need CA atom of current residue for chi1 and chi2		*/
	/* ------------------------------------------------------------ */

	if ((dchi1 || dchi2) && (! get_named_atom("CA", r, ra + 1))) {
	   if (verbose)
		   angle_error(r->name, r->num, NULL, NULL, 0, HAS_NO_CA_ATOM, 1);
	   dchi1 = dchi2 = 0;
	}

	CHECK_ANGLES;

	/* ------------------------------------------------------------ */
	/* we need CB atom of current residue for chi1, chi2 and chi3	*/
	/* ------------------------------------------------------------ */

	if ((dchi1 || dchi2 || dchi3) &&
		(! get_named_atom("CB", r, ra + 2))) {
	   if (verbose)
		   angle_error(r->name, r->num, NULL, NULL, 0, HAS_NO_CB_ATOM, 1);
	   dchi1 = dchi2 = dchi3 = 0;
	}

	CHECK_ANGLES;

	/* ------------------------------------------------------------ */
	/* we need CG(1) or OG(1) or SG atom of current residue 	*/
	/* for chi1, chi2 and chi3					*/
	/* ------------------------------------------------------------ */

	if ((dchi1 || dchi2 || dchi3) && 
	      (!( get_named_atom("CG" , r, ra + 3) /* ASP, ASN, TYR,	*/
						   /* PHE, GLU, GLN,	*/
						   /* LYS, LEU, TRP,	*/
						   /* PRO, MET, HIS	*/
	       || get_named_atom("CG1", r, ra + 3) /* VAL,ILE		*/
	       || get_named_atom("OG" , r, ra + 3) /* SER		*/
	       || get_named_atom("OG1", r, ra + 3) /* THR		*/
	       || get_named_atom("SG" , r, ra + 3) /* CYS		*/
	       ))) {
	   if (verbose)
		   angle_error(r->name, r->num, NULL, NULL, 0, HAS_NO_CG_ATOM, 1);
	   dchi1 = dchi2 = dchi3 = 0;
	}

	CHECK_ANGLES;

	/* ------------------------------------------------------------ */
	/* we need CD(1) or equivalent atom of current residue 		*/
	/* for chi2 and chi3						*/
	/* ------------------------------------------------------------ */

	if ( (dchi2 || dchi3) &&
	      (!( get_named_atom("CD" , r, ra + 4)
	       || get_named_atom("CD1", r, ra + 4)
	       || get_named_atom("OD1", r, ra + 4) 	/* ASP, ASN 	*/
	       || get_named_atom("SD" , r, ra + 4) 	/* MET		*/
	       || get_named_atom("ND1", r, ra + 4) 	/* HIS 		*/
	       ))) {
	   if (verbose)
		   angle_error(r->name, r->num, NULL, NULL, 0, HAS_NO_CD_ATOM, 1);
	   dchi2 = dchi3 = 0;
	}

	CHECK_ANGLES;

	/* ------------------------------------------------------------ */
	/* we need CE(1) or equivalent atom of current residue 		*/
	/* for chi3							*/
	/* ------------------------------------------------------------ */

	if (dchi3 &&
	      (!( get_named_atom("CE" , r, ra + 5)
	       || get_named_atom("CE1", r, ra + 5)
	       || get_named_atom("OE1", r, ra + 5) 	/* GLU,GLN	*/
	       || get_named_atom("N"  , r, ra + 5) 	/* PRO		*/
	       || get_named_atom("NE" , r, ra + 5) 	/* ARG		*/
	       || get_named_atom("NE1", r, ra + 5) 	/* TRP		*/
	       ))) {
	   if (verbose)
		   angle_error(r->name, r->num, NULL, NULL, 0, HAS_NO_CE_ATOM, 1);
	   dchi3 = 0;
	}

	CHECK_ANGLES;

	return compute_chis(r, ra, dchi1, dchi2, dchi3);

}

#undef CHECK_ANGLES

/* -------------------------------------------------------------------- */
/* computes alpha dihed. angle and tau angle				*/
/* put result in r and returns 1 if all angles defined			*/	
/* atoms coordinates are given in ra (see below)			*/	
/* -------------------------------------------------------------------- */
static int compute_alpha_tau(Residu  *r, ResAtom *ra,
		int do_tau, int do_alpha)
{
	r->tau   = (do_tau   ? atom_angle(ra)  : NO_ANGLE);
	r->alpha = (do_alpha ? atom_dihed(ra)  : NO_ANGLE);

	return (    (r->tau != NO_ANGLE)
		 && (r->alpha != NO_ANGLE));
}

/* -------------------------------------------------------------------- */
/* computes the tau angle between CA-1,CA and CA+1			*/
/* and computes the alpha angle between CA-1,CA, CA+1 and CA+2		*/
/* returns what compute_alpha_tau returns				*/
/* find an angle : 1, cannot find an angle : 0				*/
/* -------------------------------------------------------------------- */
int	get_alpha_tau(Residu *r, int verbose)
{
	int	dtau, dalpha;		/* computability flags		*/	
	ResAtom ra[4];			/* ra[0] -> CA-1		*/
					/* ra[1] -> CA  		*/
					/* ra[2] -> CA+1  		*/
					/* ra[3] -> CA+2  		*/

#define CHECK_ANGLES 	if (! (dtau || dalpha )) \
			   return compute_alpha_tau(r, ra, 0, 0)

	dtau = dalpha = 1;		/* assume all is ok		*/

	/* ------------------------------------------------------------ */
	/* verify that all residues exist				*/
	/* ------------------------------------------------------------ */
	if ( ! r) {
	   MyError(stderr, 0, "get_alpha_tau:: residue does not exist");
	   return 0;
	}
	else if ( ! (r->prev && (r->prev->num == r->num-1))) {
	   if (verbose)
		   angle_error(r->name, r->num, NULL, NULL, 0, AT_START_OF_CHAIN, 2);
	   dtau = dalpha = 0;
	}
	else if ( ! (r->next && (r->next->num == r->num+1))) {
	   if (verbose)
		   angle_error(r->name, r->num, NULL, NULL, 0, AT_END_OF_CHAIN, 2);
	   dtau = dalpha = 0;
	}
	else if(! ((r->next->next) &&
			(r->next->next->num == r->next->num+1))) {
	   if (verbose)
		   angle_error(r->next->name, r->next->num, NULL, NULL, 0,
	   					AT_END_OF_CHAIN, 2);
	   dalpha = 0;
	}

	CHECK_ANGLES;

	/* here dtau is 1 */

	/* ------------------------------------------------------------ */
	/* we need CA atom of current residue for both tau and alpha	*/
	/* ------------------------------------------------------------ */
	if ( ! get_named_atom( "CA", r, ra+1) ) {
	   if (verbose)
		   angle_error(r->name, r->num, NULL, NULL, 0, HAS_NO_CA_ATOM, 2);
	   dtau = dalpha = 0;
	}

	CHECK_ANGLES;

	/* ------------------------------------------------------------ */
	/* we need CA atom of previous residue for both tau and alpha	*/
	/* ------------------------------------------------------------ */

	if ( ! get_named_atom( "CA", r->prev, ra)) {
	   if (verbose)
		   angle_error(r->prev->name, r->prev->num, NULL, NULL, 0,
	   						HAS_NO_CA_ATOM, 2);
	   dtau = dalpha = 0;
	}

	CHECK_ANGLES;

	
	/* ------------------------------------------------------------ */
	/* we need CA atom of next residue for both tau and alpha	*/
	/* ------------------------------------------------------------ */
	if ( ! get_named_atom( "CA", r->next, ra+2)) {
	   angle_error(r->next->name, r->next->num, NULL, NULL, 0,
	   					HAS_NO_CA_ATOM, 2);
	   dtau = dalpha = 0;
	}

	CHECK_ANGLES;

	/* here dtau is 1 */

	/* ------------------------------------------------------------ */
	/* we need CA atom of next->next residue for alpha		*/
	/* ------------------------------------------------------------ */

	if (dalpha && (! get_named_atom( "CA", r->next->next, ra+3))) {
	   if (verbose)
		   angle_error(r->next->next->name, r->next->next->num,
	   				NULL, NULL, 0, HAS_NO_CA_ATOM, 2);
	   dalpha = 0;
	}

	return compute_alpha_tau(r, ra, dtau, dalpha);

#undef CHECK_ANGLES


}

/* -------------------------------------------------------------------- */
/* extract a pdb field from a pdb line					*/
/* field starts at start and has a length len				*/
/* -------------------------------------------------------------------- */
static char *pdb_field(char *line, int start, int len)
{
	static char buffer[80];

	strncpy(buffer, line + start, len);
	buffer[len] = '\000';

	return buffer;
}

/* -------------------------------------------------------------------- */
/* read an atom (in a pdb line)						*/
/* -------------------------------------------------------------------- */

#define GET_FIELD(start, len, fmt, var) 		  		\
	if (sscanf(pdb_field(line, start, len), fmt, var) != 1) {	\
	   MyError(stderr, 0, "get_atom: bad atom format");		\
	   return 0;							\
	}

#define GET_OPTFIELD(start, len, fmt, var) 		  		\
	if (sscanf(pdb_field(line, start, len), fmt, var) != 1) {	\
	   *(var) = 1.0;						\
	}

int get_atom(char *line, Atom *atom)
{
	GET_FIELD( 6, 5, "%hd", &(atom->num));

	GET_FIELD(12, 4, "%s",  atom->name);
	
	GET_FIELD(17, 3, "%s",  atom->resn);

	/* on saute un caractere apres le residu	*/
	/* ce caractere indique la chaine		*/

	if (sscanf(pdb_field(line, 21, 1), "%c", &(atom->chain)) != 1)
	   atom->chain = NO_CHAIN;

	GET_FIELD(22, 4, "%hd",  &(atom->ires));

	if (sscanf(pdb_field(line, 26, 1), "%c", &(atom->inser)) != 1)
	   atom->inser = NO_INSER;

	GET_FIELD(30, 8, "%f",   &(atom->x));
	GET_FIELD(38, 8, "%f",   &(atom->y));
	GET_FIELD(46, 8, "%f",   &(atom->z));

	GET_OPTFIELD(54, 6, "%f",   &(atom->localize));
	GET_OPTFIELD(61, 6, "%f",   &(atom->tfactor));

/*ATOM   1081  OE2 GLU B  18      -3.301  48.402  56.383  1.00 44.61           O */
/* 	fprintf(stderr, "%s\n", line); */
/* 	fprintf(stderr, "get_atom:: ATOM tFactor %f\n", atom->tfactor); */

	return 1;
}

#undef GET_FIELD

/* -------------------------------------------------------------------- */
/* reads residues and atoms from a pdb file return the first residue	*/
/* which is a dummy one	or NULL if fails				*/
/* -------------------------------------------------------------------- */
Residu	*readPdbResidues(FILE *f, char *name, int *nbres)
{
#define DESCLEN	1024

   char		line[LINE_LENGTH], key[LINE_LENGTH], 
   		desc[DESCLEN], pdbname[5];
   float	resolution;
   Atom		atom;
   Residu	*cur_res, *prev_res, *first_res;
   int		ter = 0;
 
						/* first res is a 	*/
						/* dummy res		*/
   if ((first_res = cur_res = new_residu()) == NULL)
   	return NULL;
   first_res->prev  = NULL;
   first_res->num = -32000;			/* improbable		*/

   *desc = '\000';
   if (name)
     strcpy(name,"no_name");


   resolution   =  0;
   *nbres = 0;
   
						/* -- read loop		 */
   while (fgets(line, LINE_LENGTH, f)) {


        if (line[strlen(line)-1] == '\n')
		line[strlen(line)-1] = '\000';

	if (sscanf(line, "%s", key) != 1) {
	   fprintf(stderr, "readPdbResidues: protein %s Cannot read key in line <%s>\n", name, line);
	}

	if (! strcmp(key, "HEADER")) {		/* - HEADER		*/
	   strncpy(pdbname, line + 62, 4);	/* PDB name	*/
           if (name != NULL) {
	      strncpy(name, pdbname, 4);
	      name[4] = ' ';
	      strncpy(name + 5, line+10, 40);
	      name[45] = '\000';
	   }
	}

	else if (! strcmp(key, "DESC")) {		/* - DESC		*/
				/* DESC is an ABI special addon	*/
				/* which give description for	*/
				/* PDB file			*/
           if (name != NULL) {
	      /* fprintf(stderr,"line=<%s>\n",line); */
	      strncpy(name, pdbname, 4);
	      name[4] = ' ';
	      strncpy(name + 5, line + 10, 40);	/* description	*/
	      name[45] = '\000';
	   }
	}

	else if (! (  strcmp(key, "COMPND")
		   && strcmp(key, "SOURCE"))) {	/* - COMPBD || SOURCE	*/

	   trimr(line, 70);
	   if (strlen(desc)+strlen(line+10) < DESCLEN-1)
	      strcat(desc, line + 10);
	}

	else if (! strcmp(key, "REMARK")) {	/* - REMARK		*/

	   if (! strncmp(line + 9, "2 RESOLUTION.", 13))
		sscanf(line + 22, "%f", &resolution); 
	}

	else if (strcmp(key, "ATOM") == 0 || strcmp(key, "HETATM") == 0) {  	/* - ATOM ou HETATM		*/

           if (! get_atom(line, &atom) )
           	continue;

	   if (   (strlen(atom.resn) < 3)	/* unknown		*/
	       || (! strstr(sResiduNames, atom.resn)))
		continue;

	   if (atom.ires != cur_res->pdbnum
	           || atom.inser != cur_res->inser 
		   || atom.chain != cur_res->chain || ter == 1) {	
	           				/* residue changes i.e.	*/
						/* res num changes	*/
						/* or insertion changes	*/
					/* or an other chain can begin	*/
					/* with same number of residu	*/
					/* and/or residu name		*/
					/* e.g: pdb1jch.ent		*/
		ter = 0;
		(*nbres)++;
	
						/* creation of new res	*/
		prev_res = cur_res;
	   	cur_res = new_residu();
	   	prev_res->next = cur_res;
	   	cur_res->prev  = prev_res;

	   	cur_res->num = *nbres;
	   	strcpy(cur_res->name, atom.resn);
	   	cur_res->pdbnum  = atom.ires;
	   	cur_res->chain   = atom.chain;
	   	cur_res->inser   = atom.inser;
	   	cur_res->next    = NULL;

	   }

	   if (! put_atom_in_res(&atom, cur_res))
	   	return NULL;
#ifdef D_PRINTRES
		PrintRes(cur_res);
#endif
	}

	else if (!  strcmp(key, "ENDMDL")  || ! strcmp(key, "END")  ) {
	    break;
	}
	else if  (! strcmp(key, "TER")) { 
		/* PDB File desc: The TER record indicates the end of a list of ATOM/HETATM records for a chain.*/
		/* ne pas s'arreter sur TER qui est une fin de chaine*/
		/* mais marquer une fin de chaine	*/
            ter = 1;
	} 

   } /* end while get line */
   
   return first_res;

}

/* -------------------------------------------------------------------- */
/* reads residues and atoms from a pdb file return the first residue	*/
/* which is a dummy one	or NULL if fails				*/
/* renvoie tableau des habituel avec les cvoord PDB.. de toutes les     */
/* chaines                               				*/
/*         tChain tableau avec les debuts et fin  de toutes les chaines */
/*         *nbChain: nb de chaines                                      */
/* -------------------------------------------------------------------- */
Residu	*readPdbResiduesChain(FILE *f, tyChain **tChain, char *name, int *nbChain)
{
#define DESCLEN	1024
#define DFT_NB_CHAIN 5
   char		line[LINE_LENGTH], key[LINE_LENGTH], 
   		desc[DESCLEN], pdbname[5];
   char         oldChain=' ';
   float	resolution;
   Atom		atom;
   Residu	*cur_res, *prev_res, *first_res;
   int          maxNbChain=DFT_NB_CHAIN, nbres;
   int          model=0; /* 0 means structure is not NMR or theorical*/
                         /* if 1, keeps only first model*/
   int          continuer=0, i, ter = 0;
   
						/* first res is a 	*/
						/* dummy res		*/


   if ((first_res = cur_res = new_residu()) == NULL)
   	return NULL;
   first_res->prev  = NULL;
   first_res->num = -32000;			/* improbable		*/

   *desc = '\000';
   if (name)
     strcpy(name,"no_name");

   resolution   =  0;
   nbres = 0;

   *nbChain=0;
   oldChain='\0';
   NEWY(*tChain, tyChain, maxNbChain);
   for(i=0; i<maxNbChain; i++){
     (*tChain)[i].pChain=NULL;
     (*tChain)[i].nbRes=0;
   }
						/* -- read loop		 */
#ifdef D_PRINTRES
		  fprintf(stderr, "readPdbResiduesChain Depart: nbChain %d \n", *nbChain);
#endif
   while ( (continuer == 0) && fgets(line, LINE_LENGTH, f)) {

        if (line[strlen(line)-1] == '\n')
		line[strlen(line)-1] = '\000';

	if (sscanf(line, "%s", key) != 1) {
	   fprintf(stderr, "readPdbResiduesChain: protein %s Cannot read key in line <%s>\n",name, line);
	}

	if (! strcmp(key, "HEADER")) {		/* - HEADER		*/
	   strncpy(pdbname, line + 62, 4);	/* PDB name	*/
           if (name != NULL) {
	      strncpy(name, pdbname, 4);
	      name[4] = ' ';
	      strncpy(name + 5, line+10, 40);
	      name[45] = '\000';
	   }
	}

	else if (! strcmp(key, "DESC")) {		/* - DESC		*/
				/* DESC is an ABI special addon	*/
				/* which give description for	*/
				/* PDB file			*/
           if (name != NULL) {
	      /* fprintf(stderr,"line=<%s>\n",line); */
	      strncpy(name, pdbname, 4);
	      name[4] = ' ';
	      strncpy(name + 5, line + 10, 40);	/* description	*/
	      name[45] = '\000';
	   }
	}
	else if (! (  strcmp(key, "COMPND")
		   && strcmp(key, "SOURCE"))) {	/* - COMPBD || SOURCE	*/

	   trimr(line, 70);
	   if (strlen(desc)+strlen(line+10) < DESCLEN-1)
	      strcat(desc, line + 10);
	   
/* 	     if ( strstr(line+9, " dna")!=NULL || */
/* 		  strstr(line+9, " DNA") !=NULL || */
/* 		  strstr(line+9, " Dna") !=NULL ){ */
/* 	       fprintf(stderr, "It is not a protein structure can not compute angles\n"); */
/* 	       exit(10); *//* It is not a protein structure can not compute angles*/ 
/* 	     } */
	}

	else if (! strcmp(key, "REMARK")) {	/* - REMARK		*/

	   if (! strncmp(line + 9, "2 RESOLUTION.", 13))
               sscanf(line + 22, "%f", &resolution); 
 	}

	else if (! strcmp(key, "EXPDTA")) {	/* - EXPDTA		*/
	     if ( strstr(line+9, "NMR")!=NULL ||
		  strstr(line+9, "THEORETICAL") !=NULL)
	       model=1;
	}
	else if((model ==1) &&  (! strcmp(key, "ENDMDL"))) {
	  continuer=1;
	}
	
	else if (strcmp(key, "ATOM") == 0 || strcmp(key, "HETATM") == 0 ) {  	/* - ATOM		*/

           if (! get_atom(line, &atom) )
           	continue;

	   if (   (strlen(atom.resn) < 3)	/* unknown		*/
	       || (! strstr(sResiduNames, atom.resn)))
		continue;

	   if (atom.ires != cur_res->pdbnum
	           || atom.inser != cur_res->inser
		   || atom.chain != cur_res->chain || ter == 1) {	
	           				/* residue changes i.e.	*/
						/* res num changes	*/
						/* or insertion changes	*/
						/* or TER record before */
					/* as an other chain can begin	*/
					/* with same number of residu	*/
					/* and/or residu name		*/
					/* e.g: pdb1jch.ent		*/
		ter = 0;
		(nbres)++;
	
						/* creation of new res	*/
		prev_res = cur_res;
	   	cur_res = new_residu();
	   	prev_res->next = cur_res;
	   	cur_res->prev  = prev_res;
	   	cur_res->num = nbres;
	   	strcpy(cur_res->name, atom.resn);



	   	cur_res->pdbnum  = atom.ires;
	   	cur_res->chain   = atom.chain;
	   	cur_res->inser   = atom.inser;
	   	cur_res->next    = NULL;

		if(cur_res->chain != oldChain){ /* Chain changes*/
		  if(*nbChain != 0) {
		    (*tChain)[*nbChain-1].nbRes = nbres-1; /* update nb residus previous chain (oldChain)*/
                                                           /* -1 for current res which is not in previous chain*/
		    nbres=1;
		  }
		  if(*nbChain >= maxNbChain){/* Si trop de chaines, agrandir le tableau*/
		    REALLOC(*tChain, tyChain, maxNbChain+DFT_NB_CHAIN);
		    maxNbChain+=DFT_NB_CHAIN;
		    /*REALLOC*/
		  }
		  (*tChain)[*nbChain].pChain = cur_res;
#ifdef D_PRINTRES 
		  fprintf(stderr, "readPdbResiduesChain Change de chaine: <%c> -> <%c> (num %d), res %d\n", 
			  oldChain, cur_res->chain, *nbChain, cur_res->pdbnum);
#endif 
		  oldChain=cur_res->chain;
		  *nbChain=*nbChain+1;
		}

	   }

	   if (! put_atom_in_res(&atom, cur_res))
	   	return NULL;
/* 	   fprintf(stderr, "RES tFactor %f\n", get_CA_TFactor(cur_res, 1)); */

#ifdef D_PRINTRES
		PrintRes(cur_res);
#endif
	}

	else if (!  strcmp(key, "ENDMDL")  || ! strcmp(key, "END")  ) {
	    break;
	}
	else if  (! strcmp(key, "TER")) { 
		/* PDB File desc: The TER record indicates the end of a list of ATOM/HETATM records for a chain.*/
		/* ne pas s'arreter sur TER qui est une fin de chaine*/
		/* mais marquer une fin de chaine	*/
            ter = 1;
	} 

   } /* end while get line */

   if(*nbChain>0)
     (*tChain)[(*nbChain)-1].nbRes = nbres; /* update nb residus last chain*/   

#ifdef D_PRINTCHAIN
   PrintChain(*tChain, *nbChain);
#endif

   

   return first_res;

}


/* ----------------------------------------------------	*/
/* Lit une ligne au format de sortie de phipsi -a [-c]  */
/* retourne -1 si probleme, 0 sinon			*/
/* ----------------------------------------------------	*/
int ReadPhipsiLine(FILE *pFi, BinaryRes *pRes, int verbose)
{
  char		ligne[PATH_MAX];
  static int	first_read = 0;

			/* Lecture d'une ligne	*/

  if (fgets(ligne, 255, pFi) == NULL) {
    fprintf(stderr, "ReadPhipsiLine: no line\n");
    return -1;
  }

/* ------- exemple de formats de fichier ---------------------------------------------	*/
/* THR     1    500.0   500.0    3.86    3.94900    2.85300   30.77300 A        (cas 1) */
/* ALA    15A   500.0   500.0    3.80   22.59800   11.54300    6.32200 A        (cas 1) */
/* THR     1    500.0   500.0    3.86 A                                         (cas 2) */
/* 012345678901234567890123456789012345678901234567890123456789012345678901234567890	*/
/*           10        20        30        40        50        60        70        80	*/

  		/* on lit d'abord RESNAME RESNUM[INSER] ALPHA TAU DIST X Y Z CHAIN */

  if (sscanf(ligne, "%3s %hd%c %f %f %f %f %f %f",
                pRes->resName, &(pRes->numPDB.num[1]),
		&(pRes->numPDB.inser[0]),
		&(pRes->alpha), &(pRes->tau), &(pRes->dist),
		&(pRes->x), &(pRes->y), &(pRes->z)) == 9) {
     pRes->chain = ligne[strlen(ligne)-2];
     if (first_read == 0)
	if (verbose)
	   fprintf(stderr, "Reading input file with x y z\n");
  }
  		/* si ca ne marche pas, c'est qu'on est dans le format		*/
		/* sans xyz, c'est a dire					*/
	  	/* format RESNAME RESNUM[INSER] ALPHA TAU DIST INSER	*/
  else if (sscanf(ligne, "%3s %hd%c %f %f %f",
                pRes->resName, &(pRes->numPDB.num[1]),
		&(pRes->numPDB.inser[0]),
		&(pRes->alpha), &(pRes->tau), &(pRes->dist)) == 6) {
     pRes->chain = ligne[strlen(ligne)-2];
     if (first_read == 0)
	if (verbose)
	   fprintf(stderr, "Reading input file without x y z\n");
     pRes->x = pRes->y = pRes->z = NO_COORDINATE;
  }
  else {
     fprintf(stderr, "ReadPhipsiLine:: Bad file format in\n");
     fprintf(stderr, "<%s>\n",ligne);
     return -1;
  }

#if 0
  fprintf(stderr, "Read %4s %4d%c %10.5f %10.5f %10.5f %c\n",
  	pRes->resName, (int)pRes->numPDB.num[1],
	pRes->numPDB.inser[0],pRes->x,pRes->y,pRes->z,pRes->chain);
#endif

  first_read = 1;

  return 0;

}


/* -------------------------------------------------------------------- */
void pad_with_blank(char *h, int totsize)
{
   int	len,i;
		/* pad string with blanks until totsize-2	*/
		/* (h[totsize-1] is the final '\0'		*/
   len = (int)strlen(h);
   if (len < HEADERSIZE-1) {
      for (i = len; i < HEADERSIZE-1; i++)
	 h[i] = ' ';              /* pad with blank       */
      h[HEADERSIZE-1] = '\0';
   }
   return;
}

/* -------------------------------------------------------------------- */
/* write residues to a binary file "fout"                               */
/* if minpdbnum != MINPDBNUM or 0 AND maxpdbnum != MAXPDBNUM or 0,	*/
/* number of residues will be recomputed				*/
/* returns the number of residues output				*/
/* -------------------------------------------------------------------- */
int writeBinaryFile(FILE *fout, char *header, char *path, int nb_res_tot,
			Residu *first_res, int verbose,
		        char chain, int minpdbnum, int maxpdbnum)
{
			/* nb_res_tot is needed if minpdbnum != MINPDBNUM	*/
			/* and maxpdbnum != MAXPDBNUM, else it will be		*/
			/* recomputed						*/
    Residu	*cur_res;
    int	noutputres;
    BinaryRes	binres;
    int trou=0;
    char myheader[HEADERSIZE],mypath[PATH_MAX];
    
			/* for other call that phipsi...	*/
     if (minpdbnum == 0 && maxpdbnum == 0) {
        minpdbnum = MINPDBNUM;
	maxpdbnum = MAXPDBNUM;
     }
     strncpy(myheader, header, HEADERSIZE-1);
     myheader[HEADERSIZE-1]='\0';
     strncpy(mypath, path,PATH_MAX-1);
     mypath[PATH_MAX-1]='\0';
	 				/* binary output		*/
     pad_with_blank(myheader, HEADERSIZE); /* pad with blank	*/

			/* recount residues if necessary	*/
     if (minpdbnum != MINPDBNUM || maxpdbnum != MAXPDBNUM) {
        nb_res_tot = 0;
        for (cur_res = first_res ; cur_res ; cur_res = cur_res->next)
	   if ((chain == NOCHAIN_CHAR || chain == cur_res->chain) &&
		cur_res->pdbnum >= minpdbnum && cur_res->pdbnum <= maxpdbnum)
	      nb_res_tot++;
     }
			/* write the header 45 car + '\0'	*/
     fwrite (myheader, (size_t)sizeof(char), (size_t)HEADERSIZE, fout);
     fwrite (mypath, (size_t)sizeof(char), (size_t)PATH_MAX, fout);
     fwrite (&nb_res_tot, (size_t)sizeof(int), (size_t)1, fout);

     
     noutputres = 0;
     for (cur_res = first_res ; cur_res ; cur_res = cur_res->next) {

	if ((chain == NOCHAIN_CHAR || chain == cur_res->chain) && cur_res->pdbnum >= minpdbnum && cur_res->pdbnum <= maxpdbnum) {
   	   noutputres++;

           get_alpha_tau(cur_res, verbose);
           if (get_distca_1_2(cur_res, verbose) == 1)
                trou = 1;
           else
                trou = 0;
                
	   strncpy(binres.resName, cur_res->name, 4);
	   binres.numPDB.num[1] = cur_res->pdbnum;
	   binres.numPDB.inser[0] = cur_res->inser;
	   binres.numPDB.inser[1] = '\0';   /* dummy, never used */
           if (trou == 0){
               binres.alpha = cur_res->alpha;
	       binres.tau   = cur_res->tau;
	   }
	   else{
            binres.alpha = 500.0;
	        binres.tau   = 500.0;
	   }
	   binres.dist  = cur_res->dis12;
	   binres.chain  = cur_res->chain;
	   get_CA_xyz(cur_res, &(binres.x), &(binres.y), &(binres.z), verbose);

           /*
           fprintf(stderr, "adress =%X\n",(unsigned int)&binres);
           fprintf(stderr, "res trou=%d %f %f %f\n",trou, binres.alpha,binres.tau,binres.dist);
           fprintf(stderr, "res  %c\n", binres.numPDB.inser[0]);
           fprintf(stderr, "res  %c\n", binres.numPDB.inser[1]);
           fprintf(stderr, "res  %d \n",binres.numPDB.num[1]);
           fprintf(stderr, "res %d \n",binres.numPDB.num[0]);
           fprintf(stderr, "res %s %c %f %f %f\n",binres.resName,binres.chain,binres.x,binres.y,binres.z);
           fflush(stderr);
	   */
           fwrite (&binres, (size_t)sizeof(BinaryRes), (size_t)1, fout);
	}

     }
     
     return noutputres;
}


/* -------------------------------------------------------------------- */
/* reads residues and atoms from a binary file return the first residue	*/
/* which is a dummy one	or NULL if fails				*/
/* -------------------------------------------------------------------- */
BinaryRes *readBinary(FILE *f, char *header, char *path, int *nbres, int verbose)
{
   BinaryRes	*binres;
   int nbResVerif=0;

				/* header is HEADERSIZE car with '\0'	*/
   if (fread(header, (size_t)sizeof(char), (size_t)HEADERSIZE, f) != HEADERSIZE) {
	 verbose && fprintf(stderr,"readBinary: Cannot read header\n");
      return NULL;
   }
   if (fread(path, (size_t)sizeof(char), (size_t)PATH_MAX, f) != PATH_MAX) {
         verbose && fprintf(stderr,"readBinary: Cannot read path\n");
      return NULL;
   }
   if (fread(nbres, (size_t)sizeof(int), (size_t)1, f) != 1) {
         verbose && fprintf(stderr,"readBinary: Cannot read number of residues\n");
      return NULL;
   }

   if (! (binres = (BinaryRes *)malloc(sizeof(BinaryRes)*(size_t)(*nbres)))) {
         verbose && fprintf(stderr,"readBinary: Not enough memory for binres array !\n");
      return NULL;
   }

   nbResVerif=fread(binres, (size_t)sizeof(BinaryRes), (size_t)(*nbres), f);
   if (	nbResVerif != (*nbres)) {
      free(binres);
      verbose && fprintf(stderr,"readBinary: Cannot read residues, %d residues expected,  %d residus read!\n", 
			 *nbres, nbResVerif);
      return NULL;
   }
   
   return binres;

}

