/*==================================================================*/
/* Lecture des fichiers d'angles alpha (ascii ou binaire, banque ou */
/* ne contenant qu'une proteine, avec ou sans les coordonnees des   */
/* carbones alpha) et "discretisation" des angles                   */
/* Fonctions de lecture:                                            */
/*     tyProt* LitProtAlpha(FILE *pFi, int verbose);                */
/*     tyProt* BinaryReadProtAlpha(FILE *pFi, int verbose);         */
/*     static int NextProt(FILE *pFi);                              */
/*     static int LitEnTete (FILE *pFi, char *header, char *path);  */
/*     int RecodeAngles(tyProt* pProt, int angleMax);               */
/* Allouer la memoire pour les structures tyProt                    */
/*(donnees sur les proteines, cf. yakusa.h) :                       */
/*     static tyProt* NewProt(int nbRes);                           */
/* Liberer la memoire des structures tyProt :                       */
/*     void FreeProt(tyProt* pProt);                                */
/* Fichier nettoye le 9/03/2003, non anglicise                      */
/*==================================================================*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>              /* for errno                    */

#include "swelfe.h"
#include "lecture.h"
#include "util.h"
#include "libfasta.h"
#include "libabiutils.h"

#ifndef MYFREE
#define MYFREE(x)	if ((x)) free(x)
#endif

#define NB_MAX_FRAG 50
#define SIZE_NUM_PDB 7 /*num jusq 10000, + 1 lettre, + \0*/


/*tyProt* NewProt(int nbRes);*/

static int NextProt(FILE *pFi);
static int LitEnTete (FILE *pFi, char *header, char *path);

static char aa3to1(char *resname);



/*****************************************************************/
/*Read protein alpha angles with pdb format                      */
/* pFiPath must be the path of pFi				 */
/*open a temp alpha file and call BinaryReadProtAlpha            */
/*Return a tyProt structure                                      */
/*****************************************************************/
tyProt* LitPdbProtAlpha(FILE *pFi, char *pFiPath, int verbose)
{
  FILE		*tmpfil = NULL ;
  char		header[256];
  Residu	*first_res;
  int 		nb_res_tot;
  tyProt	*pProt;
  

  if (! (tmpfil = tmpfile())) {
	fprintf(stderr, "LitPdbProtAlpha:: Error: %s\n",strerror(errno));
	return NULL;
  }

  first_res = readPdbResidues(pFi, header, &nb_res_tot);
  if (first_res->next == NULL && ! feof(pFi)) {
	fprintf(stderr, "Pb in reading PDB file : residu null\n");
	free_list_residus(first_res);
     	return NULL;
  }

/*	fprintf(stderr, "residu : %d %s %s\n", (int)(first_res->next),header, pFiPath);*/  /*correction annela*/
/*	fprintf(stderr, "residu : %d %s %s\n", (int)(first_res->next->num),header, pFiPath); */ /* recorrection jompo ;-) */
/* first_res est un dummy residu */

 				/* binary output in tmpfil	*/

  writeBinaryFile(tmpfil, header, pFiPath, nb_res_tot, first_res->next, 0, NOCHAIN_CHAR, MINPDBNUM, MAXPDBNUM);

  free_list_residus(first_res);

  rewind(tmpfil);

 				/* read tyProt in tmpfil	*/
  pProt = BinaryReadProtAlpha(tmpfil, verbose); 

  fclose(tmpfil);

  return pProt;

}


/*****************************************************************/
/*Read protein alpha angles with phipsi format                   */
/*Return a tyProt structure                                      */
/* format of each protein in the file:                           */
/*HEADER <PDB code (4 characters)> <description (desc)>          */
/*PATH <path of the original pdb file (necessary only if there is*/
/*      not the coordinates with the alpha angles)>              */
/*NBRES  <number of residues>                                    */
/*Residue_name Residue_number Alpha Tau CA_CA+1_Distance \       */
/*     Residue_name [Chain character] \                          */
/*     Residue_number_inPDB[Insertion flag]                      */    
/* The three first lines are read by the function LitEnTete      */
/*The following lines are read by ReadPhipsiLine                 */
/*For more details, cf. function ReadPhipsiLine                  */
/*First abd last alpha angles of each protein must begin with    */
/* NOT_ANGLE_PHIPSI (define as 500) as alpha angle               */
/*****************************************************************/

tyProt* LitProtAlpha(FILE *pFi, int verbose)
{
  
  char pdbpath[PATH_MAX], desc[DESC_LEN];
  int nbRes=0;
  int iProt=0;
  BinaryRes unRes;
  tyProt *pProt;
  char chainePrecedente;
  int verif;
  char *pCar=NULL;
  char nomProtTmp[PATH_MAX];


  if ((nbRes=LitEnTete(pFi, desc, pdbpath)) == PASBON) {
    if(feof(pFi) == 0)
      {
	fprintf(stderr, "Problem in residue number reading in header\n");
	verif=NextProt(pFi);
	if(verif==PASBON)
	  if (verbose)
	     fprintf(stderr, "End of file encountered while searching to read next header\n");
	return(NULL);
      }
    else
      {
	if (verbose)
           fprintf(stderr, "End of file encountered while reading header\n");
	return(NULL);
      }
  }
  else if(nbRes==0) {
      fprintf(stderr, "LitProtAlpha:: No residue to read in %s\n", pdbpath);
      if(NextProt(pFi)==PASBON)
         fprintf(stderr, "LitProtAlpha:: End of file encountered while reading\n");
      return(NULL);
  }


  if(feof(pFi)!=0) {
     fprintf(stderr, "LitProtAlpha:: Empty file\n");
     return(NULL);
  }

  if (! (pProt=NewProt(nbRes))) {
      fprintf(stderr, "Problem in allocation for pProt in LitProtAlpha\n");
      return(NULL);
  }


  pProt->nbRes=nbRes;
  RemoveSpecialChar(pdbpath);
  strcpy(pProt->pdbpath, pdbpath);
  RemoveSpecialChar(desc);
  strcpy(pProt->desc, desc);

  pCar=strrchr(pProt->pdbpath, '/');
  if(pCar==NULL)
    strcpy(pProt->nomProt, pProt->pdbpath);
  else
    strcpy(pProt->nomProt, pCar+1);

  pCar=strstr(pProt->nomProt, "pdb");
  if(pCar!=NULL && pCar == pProt->nomProt) /*Si "pdb" est au debut du nom*/
    strcpy(nomProtTmp, pCar+3);
  else
    strcpy(nomProtTmp, pProt->nomProt);

  pCar=strstr(nomProtTmp, ".ent");
  if(pCar!=NULL ) 
    *pCar='\0';
     strcpy(pProt->nomProt, nomProtTmp);

/*   fprintf(stderr, "nomProt: %s %s %s \n", pProt->nomProt, pCar, nomProtTmp); */
/*   exit(0); */

/*Lecture du premier residu*/
  if (feof(pFi)!=0)
    {
      fprintf(stderr, "Incorrect number of residues: %d read residues over %d\n",
	iProt-1, nbRes-1);
      FreeProt(pProt);
      return(NULL);
    }

  if(ReadPhipsiLine(pFi, &unRes, verbose) == -1)
	fprintf(stderr, "Problem with ReadPhipsiLine in LitProtAlpha\n");
  pProt->tNumPDB[iProt]=unRes.numPDB;
  pProt->coords[iProt].x=unRes.x;
  pProt->coords[iProt].y=unRes.y;
  pProt->coords[iProt].z=unRes.z;
  pProt->tNameRes[iProt]=aa3to1(unRes.resName);
  if (unRes.alpha==NOT_ANGLE_PHIPSI)
    {
      pProt->tAngles[iProt]=NOT_INT_ANGLE;
    }
  else
    {
      pProt->tAngles[iProt]=(int)(unRes.alpha*DIVANGLE);
    }




  if (pProt->nbFrag>=NB_MAX_FRAG) {
    fprintf(stderr,
	    "Maximum number of chains/fragments (%d) exceeded for %s\n",
	    NB_MAX_FRAG, pProt->nomProt);
    if(NextProt(pFi)==PASBON)
      fprintf(stderr, "End of file reached\n");
    FreeProt(pProt);
    return(NULL);
  }
  else
    pProt->tFrag[pProt->nbFrag].name=unRes.chain;

 chainePrecedente=unRes.chain;

/*Lecture des autres residus*/
  for (iProt=1; iProt<nbRes; iProt++) {

      if (feof(pFi)!=0) {
	  fprintf(stderr, "nombre de residus incorrect: %d residus lus sur %d\n", iProt-1, nbRes-1);
	  FreeProt(pProt);
	  return(NULL);	  
      }

      if(ReadPhipsiLine(pFi, &unRes, verbose) == -1)
	      fprintf(stderr, "pb avec ReadPhipsiLine ds LitProtAlpha\n");
      if((unRes.alpha<-180) || ((unRes.alpha>180) && (unRes.alpha!=NOT_ANGLE_PHIPSI))) {
	  fprintf(stderr, "angle %f du residu %d de la proteine %s non valide , ignore\n",
		  unRes.alpha, iProt, pProt->nomProt);
	  unRes.alpha=NOT_ANGLE_PHIPSI;
      } 
      
      pProt->tNumPDB[iProt]=unRes.numPDB;
      pProt->coords[iProt].x=unRes.x;
      pProt->coords[iProt].y=unRes.y;
      pProt->coords[iProt].z=unRes.z;
      pProt->tNameRes[iProt]=aa3to1(unRes.resName);


      if(unRes.alpha==NOT_ANGLE_PHIPSI) {
	  pProt->tAngles[iProt]=NOT_INT_ANGLE;
	  if(pProt->tFrag[pProt->nbFrag].debut!=-1) {
	      if(pProt->tFrag[pProt->nbFrag].fin==-1)  {
		      				/*si c'est la fin d'une chaine*/
		  pProt->tFrag[pProt->nbFrag].fin=iProt-1;
		  pProt->nbFrag++;
	      }
	  }
	  /*Si on n'est pas dans une chaine, juste mettre l'angle dans le tableau des angles*/
      }
      else {
	      			/*si l'angle est valide*/
	  pProt->tAngles[iProt]=(int)(unRes.alpha*DIVANGLE);

	  if(pProt->tFrag[pProt->nbFrag].debut==-1) {
	    /*Si c'est le 1er residu d'une chaine*/
	    pProt->tFrag[pProt->nbFrag].debut=iProt;
	    pProt->tFrag[pProt->nbFrag].name=unRes.chain;
	    chainePrecedente=unRes.chain;
	  }
	  else {
	    /* Ce n'est pas le 1er residu mais il		*/
	    /* peut etre le 1er d'une nouvelle chaine	*/
	    /* si le nom de chaine a change			*/
	    if(unRes.chain!=chainePrecedente) {
	      if (pProt->tFrag[pProt->nbFrag].fin==-1) {
		pProt->tFrag[pProt->nbFrag].fin=iProt-1;
		pProt->nbFrag++;

	      }
	      if (pProt->nbFrag>=NB_MAX_FRAG) {
		fprintf(stderr,
			"Maximum number of chains/fragments (%d) exceeded for %s\n",
			NB_MAX_FRAG, pProt->nomProt);
		if(NextProt(pFi)==PASBON)
			fprintf(stderr, "End of file reached\n");
		FreeProt(pProt);
		return(NULL);
	      }
	      pProt->tFrag[pProt->nbFrag].debut=iProt;
	      pProt->tFrag[pProt->nbFrag].name=unRes.chain;
	      chainePrecedente=unRes.chain;
	    }

	  }

      }

     
  } /*Fin For*/

  if((pProt->tFrag[pProt->nbFrag].debut!=-1) && (pProt->tFrag[pProt->nbFrag].fin==-1))
    {
      pProt->tFrag[pProt->nbFrag].fin=iProt-1;
      pProt->nbFrag++;
    }


      
  return(pProt);
}



/************************************************************** */
/* Lecture de la query ou d'une prot de la banque		*/
/* en format binaire						*/
/* Sinon, parfaitement identique  prcdent  (LitProtAlpha)    */
/**************************************************************	*/

tyProt* BinaryReadProtAlpha(FILE *pFi, int verbose)
{
  
  char pathProt[PATH_MAX], nomProtTmp[PATH_MAX];
  char header[HEADERSIZE]; /* defini dans libphipsi.h	*/
  int nbRes=0, iProt=0;
  BinaryRes *unResPtr;
  tyProt *pProt;
  char chainePrecedente;
  char *pCar=NULL;

  if (! (unResPtr = readBinary(pFi, header, pathProt, &nbRes, verbose))) {
     if (feof(pFi)) {
	verbose && fprintf(stderr, "End of data file\n");
	return NULL;
     }
     else {
	fprintf(stderr, "Cannot read binary file (is file really in binary format?)\n");
        return NULL;
     }
  }

  if (! (pProt=NewProt(nbRes))) {
      fprintf(stderr, "BinaryReadProtAlpha: cannot allocate for pProt\n");
      MYFREE(unResPtr);
      return(NULL);
  }
  
  pProt->nbRes=nbRes;
  RemoveSpecialChar(pathProt);
  strcpy(pProt->pdbpath, pathProt);
  RemoveSpecialChar(header);
  strcpy(pProt->desc, header);

  pCar=strrchr(pProt->pdbpath, '/');
  if(pCar==NULL)
    strcpy(pProt->nomProt, pProt->pdbpath);
  else
    strcpy(pProt->nomProt, pCar+1);

  pCar=strstr(pProt->nomProt, "pdb");
  if(pCar!=NULL && pCar == pProt->nomProt) /*Si "pdb" est au debut du nom*/
    strcpy(nomProtTmp, pCar+3);
  else
    strcpy(nomProtTmp, pProt->nomProt);

  pCar=strstr(nomProtTmp, ".ent");
  if(pCar!=NULL ) 
    *pCar='\0';
     strcpy(pProt->nomProt, nomProtTmp);
    
     /*pProt->nbFrag est initialis  0 dans NewProt*/

  chainePrecedente=' ';
  for (iProt = 0; iProt < nbRes; iProt++) {
     pProt->tNumPDB[iProt] = unResPtr[iProt].numPDB;
     pProt->coords[iProt].x = unResPtr[iProt].x;
     pProt->coords[iProt].y = unResPtr[iProt].y;
     pProt->coords[iProt].z = unResPtr[iProt].z;
     pProt->tNameRes[iProt]=aa3to1(unResPtr[iProt].resName);

     if (pProt->nbFrag>=NB_MAX_FRAG) {
       if (verbose)
	 fprintf(stderr,
		 "nb max de chaines/fragments (%d) depasse pour %s\n",
		 NB_MAX_FRAG, pProt->nomProt);
       MYFREE(unResPtr);
       FreeProt(pProt);
       return(NULL);
     }
     else
       pProt->tFrag[pProt->nbFrag].name=unResPtr[iProt].chain;	/* chaine	*/

     if((unResPtr[iProt].alpha<-180.0) || ((unResPtr[iProt].alpha>180.0) && (unResPtr[iProt].alpha!=NOT_ANGLE_PHIPSI))) {
       if(unResPtr[iProt].alpha>-180.1)
	 unResPtr[iProt].alpha=-180.0;
       if(unResPtr[iProt].alpha<180.1)
 	 unResPtr[iProt].alpha=180.0;
       if((unResPtr[iProt].alpha<-180.0) || ((unResPtr[iProt].alpha>180.0) && (unResPtr[iProt].alpha!=NOT_ANGLE_PHIPSI))) {
	 if (verbose)
	   fprintf(stderr,
		   "angle %f du residu %d de la proteine %s non valide, ignore\n",
		   unResPtr[iProt].alpha, iProt, pProt->nomProt);
       unResPtr[iProt].alpha=NOT_ANGLE_PHIPSI;
       } 
     }
     pProt->tNumPDB[iProt]= unResPtr[iProt].numPDB;
     if (unResPtr[iProt].alpha==NOT_ANGLE_PHIPSI) {
       pProt->tAngles[iProt]=NOT_INT_ANGLE;
       if (iProt != 0) {
	 if(pProt->tFrag[pProt->nbFrag].debut!=-1) {
	   if(pProt->tFrag[pProt->nbFrag].fin==-1) {
	     /*si c'est la fin d'une chaine	*/
	     pProt->tFrag[pProt->nbFrag].fin=iProt-1;
	     pProt->nbFrag++;
	   }
	 }
       }
     }
     else {				/*si l'angle est valide*/
       pProt->tAngles[iProt]= (int)(unResPtr[iProt].alpha*DIVANGLE);
       if(pProt->tFrag[pProt->nbFrag].debut==-1) {
	 /*Si c'est le 1er residu d'une chaine*/
	 
	 pProt->tFrag[pProt->nbFrag].debut=iProt;
	 pProt->tFrag[pProt->nbFrag].name=unResPtr[iProt].chain;
	 chainePrecedente=unResPtr[iProt].chain;
       }
       else {			/* Ce n'est pas le 1er residu mais il peut	*/
				/* etre le 1er d'une nouvelle chaine*/
	       			/* si le nom de chaine a change*/
	 if(unResPtr[iProt].chain!=chainePrecedente) {
	   if (pProt->tFrag[pProt->nbFrag].fin==-1) {
	     pProt->tFrag[pProt->nbFrag].fin=iProt-1;
	     pProt->nbFrag++;

	   }
	   if (pProt->nbFrag>=NB_MAX_FRAG) {
	     if (verbose)
		fprintf(stderr,
			"nb max de chaines/fragments (%d) depasse pour %s\n",
			NB_MAX_FRAG, pProt->nomProt);
             MYFREE(unResPtr);
	     FreeProt(pProt);
	     return(NULL);
	   }
	   pProt->tFrag[pProt->nbFrag].debut=iProt;
	   pProt->tFrag[pProt->nbFrag].name=unResPtr[iProt].chain;
	   chainePrecedente=unResPtr[iProt].chain;
	 }
       }
     }
     if(pProt->tFrag[pProt->nbFrag].debut == 0)
       if (verbose)
	   fprintf(stderr, "BinaryReadProtAlpha:: debut frag  0 (frag %d)\n", pProt->nbFrag);
  
  }		/* fin for	*/


  if((pProt->tFrag[pProt->nbFrag].debut!=-1) && (pProt->tFrag[pProt->nbFrag].fin==-1)) {
    /*       fprintf(stderr, "fin de chaine en %d\n", iProt-1); */
    pProt->tFrag[pProt->nbFrag].fin=iProt-1;
    pProt->nbFrag++;
  }

  MYFREE(unResPtr);

  return(pProt);
}

/************************************************/
/* Allocation de la memoire necessaire          */
/* pour une structure tyProt et initialisation  */
/************************************************/
tyProt* NewProt(int nbRes)
{
  tyProt* pProt;
  int i;

  pProt=(tyProt*)malloc(sizeof(tyProt));
  if (pProt==NULL)
    {
      fprintf(stderr, "NewProt pb de malloc");
      return (NULL);
    }

  
  pProt->tAngles=(int *)malloc(nbRes*sizeof(int));
  pProt->coords=(COORDS *)malloc(nbRes*sizeof(COORDS));
  pProt->tAnglesRecodes=(int *)malloc(nbRes*sizeof(int));
  pProt->tFrag=(tyFrag *)malloc(NB_MAX_FRAG*sizeof(tyFrag));
  pProt->tNameRes=(char *)malloc((nbRes+1)*sizeof(char));  /*pour \0 decode / encode*/ 

  pProt->tNumPDB=(PdbNum *)malloc(nbRes*sizeof(PdbNum));


/*   pProt->tNumPDB=(char**)malloc(nbRes*sizeof(char*)); */
/*   if (pProt->tNumPDB==NULL){ */
/*       fprintf(stderr, "NewProt pb de malloc"); */
/*       FreeProt(pProt); */
/*       return (NULL); */
/*   } */
/*   for (i=0; i<nbRes: i++){ */
/*     pProt->tNumPDB[i]=(char**)malloc(SIZE_NUM_PDB*sizeof(char)); */
/*     if(pProt->tNumPDB[i]==NULL){ */
/*       fprintf(stderr, "NewProt pb de malloc"); */
/*       FreeProt(pProt); */
/*       return (NULL); */
/*     } */
/*   } */
    

  if ((pProt->tNumPDB==NULL) ||(pProt->tAngles==NULL) ||(pProt->tAnglesRecodes==NULL)
		  ||(pProt->tFrag==NULL) || (pProt->coords==NULL) || (pProt->tNameRes==NULL)) {
      fprintf(stderr, "NewProt pb de malloc");
      FreeProt(pProt);
      return (NULL);
  }

 for (i=0; i<nbRes; i++)
   {
     pProt->tNumPDB[i].num[1]=-1;
     pProt->tNumPDB[i].inser[0]=NO_INSER;
     pProt->tAngles[i]=-1;
     pProt->coords[i].x = pProt->coords[i].y = pProt->coords[i].z = NO_COORDINATE;
     pProt->tAnglesRecodes[i]=-1;
   }
 for (i=0; i<NB_MAX_FRAG; i++)
   {
     pProt->tFrag[i].debut=-1;
     pProt->tFrag[i].fin=-1;
     pProt->tFrag[i].name=' ';
  }
 pProt->nbFrag=0;
 pProt->nbRes=nbRes;
 pProt->pdbpath[0]='\000';
 pProt->desc[0]='\000';
 return(pProt);

}

/****************************************/
/* Libration de l'espace mmoire alou */
/* pour une structure tyProt            */
/****************************************/
void FreeProt(tyProt* pProt)
{

  if (pProt==NULL)
    return;

  MYFREE(pProt->tNumPDB);

  MYFREE(pProt->tAngles);
  MYFREE(pProt->coords);
  MYFREE(pProt->tAnglesRecodes);
 
  MYFREE(pProt->tNameRes);
  MYFREE(pProt->tFrag);

  MYFREE(pProt);

}

/********************************************************/
/*LitEnTete: Lit les 3 premires lignes de la sortie     */
/* de phipsi -p, cad :                                  */
/*HEADER <PDB code (4 characters)> <description (desc)> */
/*PATH <path of the original pdb file (necessary only if*/
/* there is not the coordinates with the alpha angles)> */
/*NBRES  <number of residues>                           */
/* Renvoie le nombre de residus                         */
/* Mais aussi en sortie:                                */
/*       path contient le path du PDB                   */
/*       header contient le code pdb et la description  */
/********************************************************/

static int LitEnTete (FILE *pFi, char *header, char *path)
{
  char ligne[256];
  int nb = 0;
  char *s;

  if(header)
    strcpy(header, "NO HEADER");

			/* read HEADER(description)	*/
  if (fgets(ligne, 255, pFi) == NULL) {
     return(PASBON); 
  }
  else if (! strncmp(ligne, "HEADER", 6)) {
    if (ligne[strlen(ligne)-1] == '\n')
       ligne[strlen(ligne)-1] = '\000';
    s = ligne;
    while (*s == ' ')
       s++;
    if (header) {
       strncpy(header, s, DESC_LEN-1);
       header[DESC_LEN-1] = '\000';
    }

  }
  else {
     fprintf(stderr, "LitEnTete: pas de HEADER sur ligne [%s]\n", ligne);
     return(PASBON);
  }

			/* read path	*/
  if (fgets(ligne, 255, pFi) == NULL) {
     if (feof(pFi)!=0)
        fprintf(stderr, "LitEnTete: Fichier vide\n");
     return(PASBON); 
  }
  else if (! strncmp(ligne, "PATH", 4)) {
    if (ligne[strlen(ligne)-1] == '\n')
       ligne[strlen(ligne)-1] = '\000';
    s = ligne + 4;
    while (*s == ' ')
       s++;
    if(path)
       strcpy(path, s);
  }
  else {
     fprintf(stderr, "LitEnTete: pas de PATH sur ligne [%s]\n", ligne);
     return(PASBON);
  }

			/* read NBRES	*/
  if (fgets(ligne, 255, pFi) == NULL) {
     if (feof(pFi)!=0)
        fprintf(stderr, "LitEnTete: Fichier vide\n");
     return(PASBON); 
  }
  else if (! strncmp(ligne, "NBRES", 5)) {
    if(sscanf(ligne+5,"%d",&nb) != 1) {
        fprintf(stderr, "LitEnTete: Cannot read number of residues\n");
        return(PASBON);
    }
  }
  else {
     fprintf(stderr, "LitEnTete: pas de NBRES sur ligne [%s]\n", ligne);
     return(PASBON);
  }

  if (nb == 0)
    return(PASBON);

  return(nb);


}


/**************************************************************/
/*Recode les angles 
On obtient des val entre 0 et ANGLEMAX (exclu)
Renvoie le tableau contenant les angles recodes (malloc a liberer)
*/
/*******************************************************************/
int RecodeAngles(tyProt* pProt, int angleMax)
{

  
  int i;
  float div;

  div=360/(float)angleMax;
  for (i=0; i<pProt->nbRes; i++)
    {
  /*Recoder les angles: [-180,+180] -> [0,36] avec par ex [5,15[ -> 1, [-175,5[ -> 0*/
      if(( (pProt->tAngles[i]>180.1*DIVANGLE) && (pProt->tAngles[i]!=NOT_INT_ANGLE)) || (pProt->tAngles[i]<-180.1*DIVANGLE))
	{
	  fprintf(stderr, "angle non valide: %f (%d)\n",
		(float)pProt->tAngles[i]/DIVANGLE, i);
	  return(PASBON);
	}

      if(pProt->tAngles[i]==NOT_INT_ANGLE)
	{
	  pProt->tAnglesRecodes[i]=angleMax;
	}
      else
	{	
	  pProt->tAnglesRecodes[i]=(int)floor((pProt->tAngles[i]/(float)DIVANGLE+
		(float)180)/div);

	  if (pProt->tAnglesRecodes[i]==angleMax) pProt->tAnglesRecodes[i]=0;
	  if(pProt->tAnglesRecodes[i]>=angleMax){
	     fprintf(stderr, "RecodeAngles:: pb de recodage, angle (%f) recode (%d) est > au maximum (%d)\n",
		     pProt->tAngles[i]/(float)DIVANGLE,pProt->tAnglesRecodes[i], angleMax); 
	     return(PASBON);
	  }
	}

    }
  return(BON);

}

/*****************************************/
/*Se positionner sur la prochaine        */
/*proteine du fichier                    */
/*(Utile lors d'erreur de lecture        */
/*comme nb max de fragments depasses)    */
/*                                       */
/*****************************************/
static int NextProt(FILE *pFi)
{
  char ligne[256];
  long	offset;

  
  offset = ftell(pFi);

  while(fgets(ligne, 255, pFi) != NULL) {
    if ( ! (strncmp(ligne, "HEADER",6))) {
        fseek (pFi, offset, SEEK_SET);
	return(BON);
    }
    offset = ftell(pFi);
  }

  return(PASBON);

}

/* -------------------------------------------------	*/
/* returns the 1 aa character code			*/
/* -------------------------------------------------	*/
static char aa3to1(char *resname)
{
    static char sAcidesAminesPdb[85] = "ALA ARG ASN ASP CYS GLN GLU GLY HIS ILE LEU LYS MET PHE PRO SER THR TRP TYR VAL ZZZ\0";
    static char AA1[22] = "ARNDCQEGHILKMFPSTWYVZ\0";
    char *pos;
    int i;

    pos = strstr(sAcidesAminesPdb, resname);
    if (pos == NULL) /*acide amin inconnu*/
      {
        return('Z');
      }
    else
      {
        i=(int)(pos-sAcidesAminesPdb)/4; /*adresse 1ere case tableau - case trouvee => nombre de case entre 2, /4 => indice dans ke tableau*/
        if (i>=20)
          {
            fprintf(stderr, "pb dans aa3to1: %s, pos: %s, cases %d indice %d\n",
                    resname, pos, (int)(pos-sAcidesAminesPdb), (int)(pos-sAcidesAminesPdb)/4);
            return('Z');
          }
        return(AA1[i]);
      }
    
}


/*****************************************************************/
/*Read protein sequence from fasta format                        */
/*Return a tyProt structure                                      */
/*****************************************************************/
/* ---------------------------------------------------- */
/* replace bad symbols in string			*/
/* ---------------------------------------------------- */
int ReplaceInStr(char *str, char *alphabet, int repchar)
{
    char *s;
    int nb  = 0;
    for (s = str ; *s ; s++)
	if (! strchr(alphabet, *s)){
            /* fprintf(stderr,"%c unknown in %s\n",*s, alphabet);  */
	    *s = repchar;
		nb ++ ;
		}
		
    return nb;
}

/* -------------------------------------------------------------------	*/
/*  indice la sequence selon l'alphabet               			*/
/* -------------------------------------------------------------------	*/
void sIndexSequence(char *seq, int *seq_recodee, int longueur)
{
    int i;
    for (i=0 ; i<longueur ; i++){
      		seq_recodee[i] =  (int) (seq[i] - 'A');
    }
}

/* -------------------------------------------------------------------	*/
/* read Fasta format in tyProt, set index of characters in sequence	*/
/*  in tyProt tAngles based upon their rank in alphabet			*/
/* -------------------------------------------------------------------	*/
tyProt* LitFormatFasta (FILE *pFi, int verbose , char *alphabet, int num)
{
  tyProt *pProt;
  FastaSequencePtr temposeq;
  int nbpasbon;

  temposeq = NewFastaSequence();
  if (temposeq==NULL) {
    fprintf(stderr,"not enought memory for temposeq (fasta)  ..\n");
    return NULL;
  }
  
  if (  (ReadFastaSequence(pFi, temposeq, num)) == Faux){
    fprintf(stderr,"IO error 1 with sequence #%s\n", temposeq->name);
    FreeFastaSequence(temposeq);
    return NULL;   
  }


  if (! temposeq->ok) {
    fprintf(stderr,"IO error 2 with sequence #%s (flag ok 0)\n", temposeq->name);
    FreeFastaSequence(temposeq);
    return NULL;
  }
  
  (void) UpperStr(temposeq->seq);
  if(strlen(alphabet) < 5) { /*ADN */
	  nbpasbon =  ReplaceInStr(temposeq->seq, alphabet,'A');
	  if((temposeq->length)%3 != 0) {
                fprintf(stderr,"sequence non multiple de 3 !!!  : %d\n", temposeq->length);
  		FreeFastaSequence(temposeq);
	  	return NULL;
	  }
	  if(nbpasbon > NBPASBON_MAX) {
                fprintf(stderr,"Too much undefined characters in #%s\n", temposeq->name);
                
  		FreeFastaSequence(temposeq);
	  	return NULL;
	  }
	  
  }
  else
	  (void) ReplaceInStr(temposeq->seq, alphabet,'X');
 
  pProt = NewProt(temposeq->length);
  if (pProt==NULL) {
    fprintf(stderr,"not enought memory for pProt (fasta)  ..\n");
    FreeFastaSequence(temposeq);
    return NULL;
  }
  
  (void)strncpy(pProt->nomProt, temposeq->name, PATH_MAX);
  pProt->nomProt[PATH_MAX-1] = '\0';
  (void)strncpy(pProt->desc, temposeq->comment,DESC_LEN );
  pProt->desc[DESC_LEN-1] = '\0';
  pProt->nbRes =  (int)temposeq->length;     
  (void)strcpy(pProt->tNameRes, temposeq->seq); 
  pProt->tNameRes[pProt->nbRes] = '\0'; 
  sIndexSequence(temposeq->seq, pProt->tAngles,temposeq->length) ; 

  FreeFastaSequence(temposeq);
  
  return pProt;
  
}


/********************            lit format fasta en binaire         ***********************/

tyProt* LitFormatFasta_Binaire (FILE *pFi, int verbose , char *alphabet, int num)
{
  tyProt *pProt;
  int nbpasbon;
  FastaSequencePtr temposeq;


#if 0
  int nbitems;
  size_t taille_seq;
  size_t taille_struct=sizeof(FastaSequence);

  temposeq = NewFastaSequence();
  if (temposeq==NULL) {
    fprintf(stderr,"not enought memory for temposeq (fasta)  ..\n");
    return NULL;
  }

  /*lecture en binaire*/
    if ((nbitems = fread(temposeq, (size_t)(taille_struct),1, pFi)) <1){
         if(feof(pFi)){ /*fin du fichier*/
            FreeFastaSequence(temposeq);
            return NULL;   
           }
    else {
            fprintf(stderr,"1 IO error 1 with sequence #%s\n", temposeq->name);
            FreeFastaSequence(temposeq);
            return NULL;   
            }
        }

        taille_seq = (size_t)(temposeq->length);

         NEWY(temposeq->seq,char, taille_seq+1);



        if((nbitems = fread(temposeq->seq, (size_t)sizeof(char), (size_t)(taille_seq),pFi)) < (size_t)(taille_seq)){

            fprintf(stderr,"2 IO error 1 with sequence #%s %d \n", temposeq->name, nbitems);
            FreeFastaSequence(temposeq);

            return NULL;   
        }

        temposeq->seq[taille_seq] = '\0';
   /*   fprintf(stderr,"seq : %s \n%s lg %d %d \n", temposeq->name, temposeq->seq, temposeq->length, taille_seq);*/

/*verification des sequences*/

#endif

/*lecture en binaire : */
  
  temposeq = Binaire2Fasta (pFi);
  if (! temposeq) {
    fprintf(stderr,"Fin de ficher \n");
    return NULL;
  }
  else if (! temposeq->ok) {
    fprintf(stderr,"IO error 2 with sequence #%s (flag not ok)\n", temposeq->name);
    FreeFastaSequence(temposeq);
    return NULL;
  }

   
  (void) UpperStr(temposeq->seq);
  if(strlen(alphabet) < 5) { /*ADN */
	  nbpasbon =  ReplaceInStr(temposeq->seq, alphabet,'A');
	  if((temposeq->length)%3 != 0) {
                fprintf(stderr,"sequence non multiple de 3 !!!  : %d\n", temposeq->length);
  		FreeFastaSequence(temposeq);
	  	return NULL;
	  }
	  if(nbpasbon > NBPASBON_MAX) {
                fprintf(stderr,"Too much undefined characters in #%s\n", temposeq->name);
                
  		FreeFastaSequence(temposeq);
	  	return NULL;
	  }
	  
  }
  else
	  (void) ReplaceInStr(temposeq->seq, alphabet,'X');
 
  pProt = NewProt(temposeq->length);
  if (pProt==NULL) {
    fprintf(stderr,"not enought memory for pProt (fasta)  ..\n");
    FreeFastaSequence(temposeq);
    return NULL;
  }
  
  (void)strncpy(pProt->nomProt, temposeq->name, PATH_MAX);
  pProt->nomProt[PATH_MAX-1] = '\0';
  (void)strncpy(pProt->desc, temposeq->comment,DESC_LEN );
  pProt->desc[DESC_LEN-1] = '\0';
  pProt->nbRes =  (int)temposeq->length;     
  (void)strcpy(pProt->tNameRes, temposeq->seq); 
  pProt->tNameRes[pProt->nbRes] = '\0'; 
  sIndexSequence(temposeq->seq, pProt->tAngles,temposeq->length) ; 
  
/*      fprintf(stderr,"seq : %s \n%s lg %d\n", temposeq->name, temposeq->seq, temposeq->length);*/



  FreeFastaSequence(temposeq);
  
 
  
  return pProt;
  
}

/* --------------------------------------------------	*/
/* read a structure					*/
/* --------------------------------------------------	*/
tyProt *LitStructure(FILE *f, char *path, int verbose, int format)
{
   tyProt *p = NULL;

   switch(format) {
          case ASC_FORMAT: 
              p = LitProtAlpha(f, verbose);
              break;
          case PDB_FORMAT: 
              p = LitPdbProtAlpha(f, path, verbose);
              break;
          case BIN_FORMAT: 
              p = BinaryReadProtAlpha(f, verbose);
              break;
   }
   return p;

}

