/* ---------------------------------------------------------------- */
/* Copyright (c) Atelier de BioInformatique			    */
/* @file: nr_gamma.c						    */
/* @desc: Numerical recipes - Gamma funct			    */
/* @note: taken "as is" from NUMERICAL RECIPES			    */
/* @+     original files :					    */
/*								    */
/* @history:							    */
/* @+	    <Gloup> : Tue Oct 29 1996 : Created			    */
/* ---------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "libnr.h"


#define ITMAX 100
#define EPS 3.0e-7

#define SQRT2 1.4142135624


/* ------------------------------------ */
/* Num. Recipes error routine		*/
/* ------------------------------------ */
void nr_error(char *error_text)

{
	fprintf(stderr,"> Numerical Recipes run-time error...\n");
	fprintf(stderr,"> %s\n", error_text);
	exit(1);
}

/* ---------------------------------------------------------------- */
double nr_gammln(double xx)

{
        double x,tmp,ser;
        static double cof[6]={76.18009173,-86.50532033,24.01409822,
                              -1.231739516,0.120858003e-2,-0.536382e-5};
        int j;

        x=xx-1.0;
        tmp=x+5.5;
        tmp -= (x+0.5)*log(tmp);
        ser=1.0;
        for (j=0;j<=5;j++) {
                x += 1.0;
                ser += cof[j]/x;
        }
        return -tmp+log(2.50662827465*ser);
}

/* ---------------------------------------------------------------- */
void nr_gser(double *gamser, double a, double x, double *gln)

{
        int n;
        double sum,del,ap;

        *gln=nr_gammln(a);
        if (x <= 0.0) {
                if (x < 0.0) nr_error("x less than 0 in routine GSER");
                *gamser=0.0;
                return;
        } else {
                ap=a;
                del=sum=1.0/a;
                for (n=1;n<=ITMAX;n++) {
                        ap += 1.0;
                        del *= x/ap;
                        sum += del;
                        if (fabs(del) < fabs(sum)*EPS) {
                                *gamser=sum*exp(-x+a*log(x)-(*gln));
                                return;
                        }
                }
                nr_error("a too large, ITMAX too small in routine GSER");
                return;
        }
}

/* ---------------------------------------------------------------- */
void nr_gcf(double *gammcf, double a, double x, double *gln)

{
        int n;
        double gold=0.0,g,fac=1.0,b1=1.0;
        double b0=0.0,anf,ana,an,a1,a0=1.0;

        *gln=nr_gammln(a);
        a1=x;
        for (n=1;n<=ITMAX;n++) {
                an=(float) n;
                ana=an-a;
                a0=(a1+a0*ana)*fac;
                b0=(b1+b0*ana)*fac;
                anf=an*fac;
                a1=x*a0+anf*a1;
                b1=x*b0+anf*b1;
                if (a1) {
                        fac=1.0/a1;
                        g=b1*fac;
                        if (fabs((g-gold)/g) < EPS) {
                                *gammcf=exp(-x+a*log(x)-(*gln))*g;
                                return;
                        }
                        gold=g;
                }
        }
        nr_error("a too large, ITMAX too small in routine GCF");
}

/* ---------------------------------------------------------------- */
double nr_gammq(double a, double x)

{
        double gamser,gammcf,gln;

        if (x < 0.0 || a < 0.0) nr_error("Invalid arguments in routine GAMMQ");
#if 1
	if (a==0.0) return (double)-1.0 ;
#endif
        if (x < (a+1.0)) {
                nr_gser(&gamser,a,x,&gln);
                return 1.0-gamser;
        } else {
                nr_gcf(&gammcf,a,x,&gln);
                return gammcf;
        }
}

/***************/
/* Erf         */
/***************/



static double sErf(double x){


double gamp;
gamp = 1.0 - nr_gammq(0.5,x*x);
return ((x<0.)? -gamp : gamp);


}

/****************************/
/*  loi normale cumulative */
/**************************/


double sCumNormal(double x, double mu, double sigma){

   double xx;	
   xx = (x-mu)/sigma/SQRT2;
   return (1.0+sErf(xx))/2.0;

}


/*
 * Lower tail quantile for standard normal distribution function.
 *
 * This function returns an approximation of the inverse cumulative
 * standard normal distribution function.  I.e., given P, it returns
 * an approximation to the X satisfying P = Pr{Z <= X} where Z is a
 * random variable from the standard normal distribution.
 *
 * The algorithm uses a minimax approximation by rational functions
 * and the result has a relative error whose absolute value is less
 * than 1.15e-9.
 *
 * Author:      Peter J. Acklam
 * Time-stamp:  2002-06-09 18:45:44 +0200
 * E-mail:      jacklam@math.uio.no
 * WWW URL:     http://www.math.uio.no/~jacklam
 *
 * C implementation adapted from Peter's Perl version
 */

#include <math.h>
#include <errno.h>

/* Coefficients in rational approximations. */
static const double a[] =
{
	-3.969683028665376e+01,
	 2.209460984245205e+02,
	-2.759285104469687e+02,
	 1.383577518672690e+02,
	-3.066479806614716e+01,
	 2.506628277459239e+00
};

static const double b[] =
{
	-5.447609879822406e+01,
	 1.615858368580409e+02,
	-1.556989798598866e+02,
	 6.680131188771972e+01,
	-1.328068155288572e+01
};

static const double c[] =
{
	-7.784894002430293e-03,
	-3.223964580411365e-01,
	-2.400758277161838e+00,
	-2.549732539343734e+00,
	 4.374664141464968e+00,
	 2.938163982698783e+00
};

static const double d[] =
{
	7.784695709041462e-03,
	3.224671290700398e-01,
	2.445134137142996e+00,
	3.754408661907416e+00
};

#define LOW 0.02425
#define HIGH 0.97575

double
ltqnorm(double p)
{
	double q, r;

	errno = 0;

	if (p < 0 || p > 1)
	{
		errno = EDOM;
		return 0.0;
	}
	else if (p == 0)
	{
		errno = ERANGE;
		return -1.7976931348e+308 /* minus "infinity" */;
		/* return -HUGE_VALF */ /* minus "infinity" */;
	}
	else if (p == 1)
	{
		errno = ERANGE;
		return 1.7976931348e+308 /* "infinity" */;
		/* return HUGE_VALF */ /* "infinity" */;
	}
	else if (p < LOW)
	{
		/* Rational approximation for lower region */
		q = sqrt(-2*log(p));
		return (((((c[0]*q+c[1])*q+c[2])*q+c[3])*q+c[4])*q+c[5]) /
			((((d[0]*q+d[1])*q+d[2])*q+d[3])*q+1);
	}
	else if (p > HIGH)
	{
		/* Rational approximation for upper region */
		q  = sqrt(-2*log(1-p));
		return -(((((c[0]*q+c[1])*q+c[2])*q+c[3])*q+c[4])*q+c[5]) /
			((((d[0]*q+d[1])*q+d[2])*q+d[3])*q+1);
	}
	else
	{
		/* Rational approximation for central region */
    		q = p - 0.5;
    		r = q*q;
		return (((((a[0]*r+a[1])*r+a[2])*r+a[3])*r+a[4])*r+a[5])*q /
			(((((b[0]*r+b[1])*r+b[2])*r+b[3])*r+b[4])*r+1);
	}
}
